

import os
import gzip
import bz2
import lzma
import json




class openReadBinary(object):

	def __init__(self, filePath):
		self.__filePath = filePath
		if filePath.endswith(".gz") or filePath.endswith(".gzip"):
			self.__open = gzip.open
		elif filePath.endswith(".bz2") or filePath.endswith(".bzip2"):
			self.__open = bz2.open
		elif filePath.endswith(".lzma") or filePath.endswith(".xz"):
			self.__open = lzma.open
		else:
			self.__open = open
	#

	def __enter__(self):
		self.__f = self.__open(self.__filePath, "rb")
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
	#

#

class openReadText(object):

	class _BinaryWrapper(object):

		def __init__(self, f):
			self.__f = f
		#

		def read(self, length = None):
			data = self.__f.read()
			return data.decode("utf-8")
		#

		def close(self):
			self.__f.close()
		#

	#

	def __init__(self, filePath):
		self.__filePath = filePath
		if filePath.endswith(".gz") or filePath.endswith(".gzip"):
			self.__open = gzip.open
			self.__bToText = True
		elif filePath.endswith(".bz2") or filePath.endswith(".bzip2"):
			self.__open = bz2.open
			self.__bToText = True
		elif filePath.endswith(".lzma") or filePath.endswith(".xz"):
			self.__open = lzma.open
			self.__bToText = True
		else:
			self.__open = open
			self.__bToText = False
	#

	def __enter__(self):
		if self.__bToText:
			self.__f = openReadText._BinaryWrapper(self.__open(self.__filePath, "rb"))
		else:
			self.__f = self.__open(self.__filePath, "r")
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
	#

#

class openReadJSON(object):

	class _BinaryWrapper(object):

		def __init__(self, f):
			self.__f = f
		#

		def read(self, length = None):
			data = self.__f.read()
			return data.decode("utf-8")
		#

		def close(self):
			self.__f.close()
		#

	#

	class _JSONWrapper(object):

		def __init__(self, f):
			self.__f = f
		#

		def read(self):
			data = self.__f.read()
			return json.loads(data)
		#

		def close(self):
			self.__f.close()
		#

	#

	def __init__(self, filePath):
		self.__filePath = filePath
		if filePath.endswith(".gz") or filePath.endswith(".gzip"):
			self.__open = gzip.open
			self.__bToText = True
		elif filePath.endswith(".bz2") or filePath.endswith(".bzip2"):
			self.__open = bz2.open
			self.__bToText = True
		elif filePath.endswith(".lzma") or filePath.endswith(".xz"):
			self.__open = lzma.open
			self.__bToText = True
		else:
			self.__open = open
			self.__bToText = False
	#

	def __enter__(self):
		if self.__bToText:
			self.__f = openReadJSON._JSONWrapper(openReadJSON._BinaryWrapper(self.__open(self.__filePath, "rb")))
		else:
			self.__f = openReadJSON._JSONWrapper(self.__open(self.__filePath, "r"))
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
	#

#





class openWriteBinary(object):

	def __init__(self, filePath, bSafeWrite:bool = True, compression:str = None):
		self.__filePath = filePath
		self.__bSafeWrite = bSafeWrite
		if self.__bSafeWrite:
			self.__filePathTmp = filePath + ".tmp"

		self.__format = None
		if compression:
			if (compression == "gzip") or (compression == "gz"):
				self.__open = gzip.open
			elif (compression == "bz2") or (compression == "bzip2"):
				self.__open = bz2.open
			elif compression == "lzma":
				self.__format = lzma.FORMAT_ALONE
				self.__open = lzma.open
			elif compression == "xz":
				self.__format = lzma.FORMAT_XZ
				self.__open = lzma.open
			else:
				raise Exception("Unknown compressor: " + compression)
		else:
			self.__open = open
	#

	def __enter__(self):
		if self.__bSafeWrite:
			if self.__format:
				self.__f = self.__open(self.__filePathTmp, "wb", format=self.__format)
			else:
				self.__f = self.__open(self.__filePathTmp, "wb")
		else:
			if self.__format:
				self.__f = self.__open(self.__filePath, "wb", format=self.__format)
			else:
				self.__f = self.__open(self.__filePath, "wb")
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
		if self.__bSafeWrite:
			if os.path.isfile(self.__filePath):
				os.unlink(self.__filePath)
			os.rename(self.__filePathTmp, self.__filePath)
	#

#



class openWriteText(object):

	#
	# Write text data to an underlying binary stream
	#
	class _BinaryWrapper(object):

		def __init__(self, f):
			self.__f = f
		#

		def write(self, textData):
			self.__f.write(textData.encode("utf-8"))
		#

		def close(self):
			self.__f.close()
		#

	#

	def __init__(self, filePath, bSafeWrite:bool = True, compression:str = None):
		self.__filePath = filePath
		self.__bSafeWrite = bSafeWrite
		if self.__bSafeWrite:
			self.__filePathTmp = filePath + ".tmp"
		self.__bToBinary = False

		self.__format = None
		if compression:
			if (compression == "gzip") or (compression == "gz"):
				self.__open = gzip.open
				self.__bToBinary = True
			elif (compression == "bz2") or (compression == "bzip2"):
				self.__open = bz2.open
				self.__bToBinary = True
			elif compression == "lzma":
				self.__format = lzma.FORMAT_ALONE
				self.__open = lzma.open
				self.__bToBinary = True
			elif compression == "xz":
				self.__format = lzma.FORMAT_XZ
				self.__open = lzma.open
				self.__bToBinary = True
			else:
				raise Exception("Unknown compressor: " + compression)
		else:
			self.__open = open
	#

	def __enter__(self):
		if self.__bSafeWrite:
			if self.__bToBinary:
				if self.__format:
					self.__f = openWriteText._BinaryWrapper(self.__open(self.__filePathTmp, "wb", format=self.__format))
				else:
					self.__f = openWriteText._BinaryWrapper(self.__open(self.__filePathTmp, "wb"))
			else:
				if self.__format:
					self.__f = self.__open(self.__filePathTmp, "w", format=self.__format)
				else:
					self.__f = self.__open(self.__filePathTmp, "w")
		else:
			if self.__bToBinary:
				if self.__format:
					self.__f = openWriteText._BinaryWrapper(self.__open(self.__filePath, "wb", format=self.__format))
				else:
					self.__f = openWriteText._BinaryWrapper(self.__open(self.__filePath, "wb"))
			else:
				if self.__format:
					self.__f = self.__open(self.__filePath, "w", format=self.__format)
				else:
					self.__f = self.__open(self.__filePath, "w")
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
		if self.__bSafeWrite:
			if os.path.isfile(self.__filePath):
				os.unlink(self.__filePath)
			os.rename(self.__filePathTmp, self.__filePath)
	#

#

class openWriteJSON(object):

	#
	# Write JSON data to an underlying text stream
	#
	class _JSONWrapper(object):

		def __init__(self, f, bPretty:bool):
			self.__f = f
			self.__bPretty = bPretty
		#

		def write(self, objData):
			assert isinstance(objData, (bool, int, float, str, tuple, list, dict, None))
			if self.__bPretty:
				self.__f.write(json.dumps(objData, indent="\t", sort_keys=True))
			else:
				self.__f.write(json.dumps(objData))
		#

		def close(self):
			self.__f.close()
		#

	#

	class _BinaryWrapper(object):

		def __init__(self, f):
			self.__f = f
		#

		def write(self, textData):
			self.__f.write(textData.encode("utf-8"))
		#

		def close(self):
			self.__f.close()
		#

	#

	def __init__(self, filePath, bSafeWrite:bool = True, compression:str = None, bPretty:bool = False):
		self.__filePath = filePath
		self.__bSafeWrite = bSafeWrite
		if self.__bSafeWrite:
			self.__filePathTmp = filePath + ".tmp"
		self.__bToBinary = False
		self.__bPretty = bPretty

		self.__format = None
		if compression:
			if (compression == "gzip") or (compression == "gz"):
				self.__open = gzip.open
				self.__bToBinary = True
			elif (compression == "bz2") or (compression == "bzip2"):
				self.__open = bz2.open
				self.__bToBinary = True
			elif compression == "lzma":
				self.__format = lzma.FORMAT_ALONE
				self.__open = lzma.open
				self.__bToBinary = True
			elif compression == "xz":
				self.__format = lzma.FORMAT_XZ
				self.__open = lzma.open
				self.__bToBinary = True
			else:
				raise Exception("Unknown compressor: " + compression)
		else:
			self.__open = open
	#

	def __enter__(self):
		if self.__bSafeWrite:
			if self.__bToBinary:
				if self.__format:
					self.__f = openWriteJSON._BinaryWrapper(self.__open(self.__filePathTmp, "wb", format=self.__format))
				else:
					self.__f = openWriteJSON._BinaryWrapper(self.__open(self.__filePathTmp, "wb"))
			else:
				if self.__format:
					self.__f = self.__open(self.__filePathTmp, "w", format=self.__format)
				else:
					self.__f = self.__open(self.__filePathTmp, "w")
		else:
			if self.__bToBinary:
				if self.__format:
					self.__f = openWriteJSON._BinaryWrapper(self.__open(self.__filePath, "wb", format=self.__format))
				else:
					self.__f = openWriteJSON._BinaryWrapper(self.__open(self.__filePath, "wb"))
			else:
				if self.__format:
					self.__f = self.__open(self.__filePath, "w", format=self.__format)
				else:
					self.__f = self.__open(self.__filePath, "w")
		self.__f = openWriteJSON._JSONWrapper(self.__f, self.__bPretty)
		return self.__f
	#

	def __exit__(self, *args):
		self.__f.close()
		if self.__bSafeWrite:
			if os.path.isfile(self.__filePath):
				os.unlink(self.__filePath)
			os.rename(self.__filePathTmp, self.__filePath)
	#

#





def writePrivateBinaryFile(binaryData, filePath:str):
	assert isinstance(filePath, str)
	assert isinstance(binaryData, (bytes, bytearray))

	try:
		os.remove(filePath)
	except OSError:
		pass

	umask_original = os.umask(0)
	try:
		fdesc = os.open(filePath, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o600)
	finally:
		os.umask(umask_original)

	with os.fdopen(fdesc, "wb") as fout:
		fout.write(binaryData)
#



def writePrivateJSONFile(jsonData, filePath:str, bPretty:bool = False):
	assert isinstance(filePath, str)
	assert isinstance(jsonData, (int, float, bool, str, list, tuple, dict))

	if bPretty:
		s = json.dumps(jsonData, indent="\t", sort_keys=True)
	else:
		s = json.dumps(jsonData)

	writePrivateBinaryFile(s.encode("utf-8"), filePath)
#



def loadBinaryFile(filePath:str):
	assert isinstance(filePath, str)

	with open(filePath, "rb") as fin:
		return fin.read()
#














from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# An atom-like token pattern.
#
class TokenPattern(AbstractTokenPattern):

	#
	# @param		str tokenType			(required) The type identifier of a token, e.g. "int"
	# @param		str tokenValue			(optional) The text value the token is expected to have for a match.
	# @param		str assignToVar			(optional) On match use the value as the string it is within the token.
	# @param		str assignToVarTyped	(optional) On match use the value but wrap the token text into `TypedValue`,
	#										using the token type as type information for `TypedValue'. Thus this way type information
	#										is preserved.
	# @param		bool bVarIsArray		(optional) This variable is an array. If it is an array, data is stacked as an array.
	#
	def __init__(self, tokenType:str, tokenValue:str = None, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		super().__init__()

		assert isinstance(tokenType, str)
		assert isinstance(tokenValue, (type(None), str))
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		self.__tokenType = tokenType
		self.__tokenValue = tokenValue
		self.__assignToVar = assignToVar
		self.__assignToVarTyped = assignToVarTyped
		self.__bVarIsArray = bVarIsArray
	#

	def _tryMatch(self, tokens:list, pos:int, stack:Stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		if pos >= len(tokens):
			return (False, 0)
		if tokens[pos].type != self.__tokenType:
			return (False, 0)
		if (self.__tokenValue != None) and (tokens[pos].value != self.__tokenValue):
			return (False, 0)
		if self.__assignToVar != None:
			stack.push(("v", tokens[pos], self.__assignToVar, self.__bVarIsArray))
		if self.__assignToVarTyped != None:
			stack.push(("vv", tokens[pos], self.__assignToVarTyped, self.__bVarIsArray))
		self._addTagsToStack(stack)
		return (True, 1)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPattern(self.__tokenType, self.__tokenValue, assignToVar, assignToVarTyped, bVarIsArray)
	#

#













import asyncio
import collections

from asyncio.coroutines import coroutine


class QueueEmpty(Exception):
	"""Exception raised when Queue.get_nowait() is called on a Queue object
	which is empty.
	"""
	pass


class QueueFull(Exception):
	"""Exception raised when the Queue.put_nowait() method is called on a Queue
	object which is full.
	"""
	pass

class QueueFinished(Exception):
	"""Exception raised when the Queue.put_nowait() method is called on a Queue
	object which is already finished.
	"""
	pass

class QueueStopped():
	pass

class Queue:
	"""A queue, useful for coordinating producer and consumer coroutines.

	If maxsize is less than or equal to zero, the queue size is infinite. If it
	is an integer greater than 0, then "yield from put()" will block when the
	queue reaches maxsize, until an item is removed by get().

	Unlike the standard library Queue, you can reliably know this Queue's size
	with qsize(), since your single-threaded asyncio application won't be
	interrupted between calling qsize() and doing an operation on the Queue.
	"""

	STOPPED = QueueStopped()

	def __init__(self, maxsize=0, *, loop=None):
		if loop is None:
			self._loop = asyncio.events.get_event_loop()
		else:
			self._loop = loop
		self._maxsize = maxsize

		# Futures.
		self._getters = collections.deque()
		# Futures.
		self._putters = collections.deque()
		self._bFinished = False
		self._finished = asyncio.locks.Event(loop=self._loop)
		self._init(maxsize)

	# These three are overridable in subclasses.

	def _init(self, maxsize):
		self._queue = collections.deque()

	def _get(self):
		return self._queue.popleft()

	def _put(self, item):
		self._queue.append(item)

	# End of the overridable methods.

	def _wakeup_next(self, waiters):
		# Wake up the next waiter (if any) that isn't cancelled.
		while waiters:
			waiter = waiters.popleft()
			if not waiter.done():
				waiter.set_result(None)
				break

	def __repr__(self):
		return '<{} at {:#x} {}>'.format(
			type(self).__name__, id(self), self._format())

	def __str__(self):
		return '<{} {}>'.format(type(self).__name__, self._format())

	def _format(self):
		result = 'maxsize={!r}'.format(self._maxsize)
		if getattr(self, '_queue', None):
			result += ' _queue={!r}'.format(list(self._queue))
		if self._getters:
			result += ' _getters[{}]'.format(len(self._getters))
		if self._putters:
			result += ' _putters[{}]'.format(len(self._putters))
		return result

	def qsize(self):
		"""Number of items in the queue."""
		return len(self._queue)

	@property
	def maxsize(self):
		"""Number of items allowed in the queue."""
		return self._maxsize

	def empty(self):
		"""Return True if the queue is empty, False otherwise."""
		return not self._queue

	def full(self):
		"""Return True if there are maxsize items in the queue.

		Note: if the Queue was initialized with maxsize=0 (the default),
		then full() is never True.
		"""
		if self._maxsize <= 0:
			return False
		else:
			return self.qsize() >= self._maxsize

	@coroutine
	def put(self, item):
		"""Put an item into the queue.

		Put an item into the queue. If the queue is full, wait until a free
		slot is available before adding item.

		This method is a coroutine.
		"""
		while self.full():
			putter = self._loop.create_future()
			self._putters.append(putter)
			try:
				yield from putter
			except:
				putter.cancel()  # Just in case putter is not done yet.
				if not self.full() and not putter.cancelled():
					# We were woken up by get_nowait(), but can't take
					# the call.  Wake up the next in line.
					self._wakeup_next(self._putters)
				raise
		return self.put_nowait(item)

	def put_nowait(self, item):
		"""Put an item into the queue without blocking.

		If no free slot is immediately available, raise QueueFull.
		"""
		if self._bFinished:
			raise QueueFinished
		if self.full():
			raise QueueFull
		self._put(item)
		self._bFinished = False
		self._finished.clear()
		self._wakeup_next(self._getters)

	@coroutine
	def get(self):
		"""Remove and return an item from the queue.

		If queue is empty, wait until an item is available.

		This method is a coroutine.
		"""
		while self.empty():
			if self._bFinished:
				return AsyncQueue.STOPPED
			getter = self._loop.create_future()
			self._getters.append(getter)
			try:
				yield from getter
			except:
				getter.cancel()  # Just in case getter is not done yet.
				if not self.empty() and not getter.cancelled():
					# We were woken up by put_nowait(), but can't take
					# the call.  Wake up the next in line.
					self._wakeup_next(self._getters)
				raise
		return self.get_nowait()

	def get_nowait(self):
		"""Remove and return an item from the queue.

		Return an item if one is immediately available, else raise QueueEmpty.
		"""
		if self.empty():
			if self._bFinished:
				return AsyncQueue.STOPPED
			raise QueueEmpty
		item = self._get()
		self._wakeup_next(self._putters)
		return item

	def completed(self):
		"""Indicate that producing data is completed.
		"""
		self._finished.set()
		self._bFinished = True

		for i in range(0, len(self._getters)):
			self._wakeup_next(self._getters)

	@coroutine
	def join(self):
		"""Block until all items in the queue have been gotten and processed.

		The count of unfinished tasks goes up whenever an item is added to the
		queue. The count goes down whenever a consumer calls task_done() to
		indicate that the item was retrieved and all work on it is complete.
		When the count of unfinished tasks drops to zero, join() unblocks.
		"""
		yield from self._finished.wait()











from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# A token pattern alternative pattern. One of the specified patterns must match.
#
class TokenPatternAlternatives(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern[] tokenPatterns			(required) A set of token patterns to match against.
	#
	def __init__(self, tokenPatterns:list):
		super().__init__()

		assert isinstance(tokenPatterns, list)
		for t in tokenPatterns:
			assert isinstance(t, AbstractTokenPattern)
		self.__tokenPatterns = tokenPatterns
		self.__len = len(tokenPatterns)
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		m = stack.mark()
		for i in range(0, self.__len):
			(bResult, n) = self.__tokenPatterns[i]._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if bResult:
				self._addTagsToStack(stack)
				return (True, n)
			m.resetToMark()
		return (False, 0)
	#

	def derive(self, assignToVar = None, assignToVarTyped = None, bVarIsArray = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		tokenPatterns = []
		for tokenPattern in self.__tokenPatterns:
			tokenPatterns.append(tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
		return TokenPatternAlternatives(tokenPatterns)
	#

#












import os
import array
import math




#
# Base class for RNGs.
#
class AbstractRandomNumberGenerator(object):

	_bits64n = int(-(65536//2*65536*65536*65536) + 1)
	_bits32n = int(-(65536//2*65536) + 1)
	_bits16n = int(-(65536//2) + 1)
	_bits8n = -127

	def nextUInt32(self) -> int:
		raise NotImplementedError()
	#

	def nextInt32(self) -> int:
		return self.nextUInt32() + AbstractRandomNumberGenerator._bits32n
	#

	def nextInt16(self) -> int:
		return self.nextUInt16() + AbstractRandomNumberGenerator._bits16n
	#

	def nextInt8(self) -> int:
		return self.nextUInt8() + AbstractRandomNumberGenerator._bits8n
	#

	def nextInt64(self) -> int:
		return self.nextUInt64() + AbstractRandomNumberGenerator._bits64n
	#

	def nextBytes(self, count:int) -> bytes:
		raise NotImplementedError()
	#

	def nextUInt8(self) -> int:
		return self.nextUInt32() & 255
	#

	def nextUInt16(self) -> int:
		return self.nextUInt32() & 65535
	#

	def nextUInt64(self) -> int:
		a = self.nextUInt32()
		b = self.nextUInt32()
		return (a << 32) | b
	#

	def nextString(self, poolOfCharacters:str, count:int):
		assert isinstance(poolOfCharacters, str)
		assert len(poolOfCharacters) > 1
		maxCharacters = len(poolOfCharacters)

		s = []
		while len(s) < count:
			s.append(poolOfCharacters[int(self.nextFloat() * maxCharacters)])

		return "".join(s)
	#

	def nextCharacter(self, poolOfCharacters:str):
		assert isinstance(poolOfCharacters, str)
		assert len(poolOfCharacters) > 1
		maxCharacters = len(poolOfCharacters)

		return poolOfCharacters[int(self.nextFloat() * maxCharacters)]
	#

	def nextUInt32Bounded(self, maxValue:int) -> int:
		assert isinstance(maxValue, int)
		assert maxValue > 1

		return int(self.nextFloat() * maxValue)
	#

	def nextFloat(self) -> float:
		raise NotImplementedError()
	#

#




#
# This class provides a linear congruential generator which is a pseudo number generator that if instantiated will always return the same random values.
# (See: https://en.wikipedia.org/wiki/Linear_congruential_generator)
# The internal values used here are the values Donald Knuth used for his MMIX.
#
class LinearCongruentialGenerator(AbstractRandomNumberGenerator):

	_bits64 = int("0xffffffffffffffff", 16)
	_bits32 = int("0xffffffff", 16)

	def __init__(self):
		super().__init__()

		self.__x = 0
		self.__a = 6364136223846793005
		self.__c = 1442695040888963407
	#

	def nextFloat(self) -> float:
		return self.nextUInt32() / LinearCongruentialGenerator._bits32
	#

	def nextUInt32(self) -> int:
		self.__x = (self.__a * self.__x + self.__c) & LinearCongruentialGenerator._bits64
		return self.__x & LinearCongruentialGenerator._bits32
	#

	def nextBytes(self, count:int) -> bytes:
		assert isinstance(count, int)
		assert count >= 1

		vals = []
		for i in range(0, (count + 3) // 4):
			self.__x = (self.__a * self.__x + self.__c) & LinearCongruentialGenerator._bits64
			v32 = self.__x & LinearCongruentialGenerator._bits32
			vals.append(v32)
		return array.array("I", vals).tobytes()[0:count]
	#

#




#
# This is a true /dev/urandom based random number generator. Please note that it will likely be quite inefficient as for every call
# to a random number generation method a <c>open()</c>, <c>read()</c> and <c>close()</c> is performed.
#
class URandomGenerator(AbstractRandomNumberGenerator):

	_bits64 = int("0xffffffffffffffff", 16)
	_bits32 = int("0xffffffff", 16)

	def __init__(self):
		super().__init__()
	#

	def nextFloat(self) -> float:
		vals = bytearray()
		with open("/dev/urandom", "rb") as f:
			while len(vals) < 8:
				vals.extend(f.read(8 - len(vals)))

		return int.from_bytes(vals, byteorder='big', signed=False) / URandomGenerator._bits64
	#

	def nextUInt32(self) -> int:
		vals = bytearray()
		with open("/dev/urandom", "rb") as f:
			while len(vals) < 4:
				vals.extend(f.read(4 - len(vals)))

		return int.from_bytes(vals, byteorder='big', signed=False)
	#

	def nextUInt64(self) -> int:
		vals = bytearray()
		with open("/dev/urandom", "rb") as f:
			while len(vals) < 8:
				vals.extend(f.read(8 - len(vals)))

		return int.from_bytes(vals, byteorder='big', signed=False)
	#

	def nextBytes(self, count:int) -> bytes:
		assert isinstance(count, int)
		assert count >= 1

		vals = bytearray()
		bytesToRead = count
		with open("/dev/urandom", "rb") as f:
			while bytesToRead > 0:
				vals.extend(f.read(bytesToRead))
				bytesToRead = count - len(vals)
		return vals
	#

#











import signal



class GracefullyHandleInterrupts(object):

	def __init__(self, log, bCatchSigInt = True, bCatchSigTerm = True,
		messageSigInt = "SIGINT received. Terminating gracefuly ...",
		messageSigTerm = "SIGINT received. Terminating gracefuly ..."):

		self.__bSignalRaised = False
		self.__signal_received = False
		self.__log = log
		self.__messageSigInt = messageSigInt
		self.__messageSigTerm = messageSigTerm
		self.__bCatchSigInt = bCatchSigInt
		self.__bCatchSigTerm = bCatchSigTerm
		self.__old_handler_sigint = None
		self.__old_handler_sigterm = None

	def __enter__(self):
		self.__bSignalRaised = False
		self.__signal_received = False
		if self.__bCatchSigInt:
			self.__old_handler_sigint = signal.getsignal(signal.SIGINT)
			signal.signal(signal.SIGINT, self.handler)
		if self.__bCatchSigTerm:
			self.__old_handler_sigterm = signal.getsignal(signal.SIGTERM)
			signal.signal(signal.SIGTERM, self.handler)

	def handler(self, sig, frame):
		self.__signal_received = (sig, frame)
		if sig == signal.SIGINT:
			if self.__messageSigInt != None:
				self.__log.info(self.__messageSigInt)
		elif sig == signal.SIGTERM:
			if self.__bCatchSigTerm != None:
				self.__log.info(self.__bCatchSigTerm)
		else:
			self.__log.error("Implementation error!")

	def __exit__(self, type, value, traceback):
		if self.__bCatchSigInt:
			signal.signal(signal.SIGINT, self.__old_handler_sigint)
		if self.__bCatchSigTerm:
			signal.signal(signal.SIGTERM, self.__old_handler_sigterm)

	@property
	def isInterrupted(self):
		return self.__signal_received != False


















from ..Stack import Stack
from ..TypedValue import TypedValue





#
# Abstract base class for all token pattern classes.
#
class AbstractTokenPattern(object):

	def __init__(self):
		self.__tags = None
	#

	#
	# Tries to match at the specified position.
	#
	# @param		list tokens				The list of tokens.
	# @param		int offset				Where to try a match.
	# @param		dict defaults			A set of defaults the return data map should be populated with.
	# @return		tuple					Returns a tuple containing the following data:
	#										* bool ---- Indicates success or failure
	#										* int ---- The number of tokens eaten, <c>0</c> if nothing could be eaten
	#										* dict ---- A dictionary containing the parsed data or <c>None</c> if nothing could be eaten
	#
	def tryMatch(self, tokens, offset = 0, defaults = None):
		assert isinstance(tokens, list)
		assert isinstance(offset, int)
		assert isinstance(defaults, (type(None), dict))

		stack = Stack()
		(bResult, n) = self._tryMatch(tokens, offset, stack)
		#print("---- tryMatch: " + str(bResult) + ", " + str(n))
		if bResult:
			ret = {
				"lineNo": tokens[offset].lineNo,
				"colNo": tokens[offset].colNo,
			}
			if defaults != None:
				ret.update(defaults)

			for (entryType, a, b, c) in stack:

				if (entryType == "v") or (entryType == "vv") or (entryType == "d"):

					# there is a value to store.

					# step 1: get the data to store
					if entryType == "d":
						# pure data
						value, varName, bVarIsArray = a, b, c
					elif entryType == "v":
						# take value as is
						token, varName, bVarIsArray = a, b, c
						value = token.value
					else:
						# add typed value
						token, varName, bVarIsArray = a, b, c
						value = TypedValue(token.type, token.value)

					# step 2: actually store the value
					v = ret.get(varName, None)
					if v is None:
						if bVarIsArray:
							ret[varName] = [ value ]
						else:
							ret[varName] = value
					else:
						if isinstance(v, list):
							v.append(value)
							ret[varName] = v
						else:
							ret[varName] = [ v, value ]

				elif entryType == "t":
					a, varName, value = a, b, c
					ret[varName] = value

				else:
					raise Exception("Implementation Error!")

			return (True, n, ret)
		else:
			return (False, 0, None)
	#

	@property
	def tags(self):
		if self.__tags != None:
			return dict(self.__tags)
		else:
			return None
	#

	#
	# Associate a key with a value at this pattern. If parsing succeeds the result data map will receive this key-value-pair.
	#
	def setTag(self, tagName, tagValue):
		if self.__tags is None:
			self.__tags = {}
		self.__tags[tagName] = tagValue
		return self
	#

	def _addTagsToStack(self, stack):
		if self.__tags != None:
			for tagName in self.__tags.keys():
				stack.push(("t", None, tagName, self.__tags[tagName]))
	#

	def _tryMatch(self, tokens, pos, stack):
		pass
	#

	#
	# Derive a pattern from this pattern with specific variations.
	#
	# @param	str assignToVar				If set the result data map will receive the value directly as it is.
	# @param	str assignToVarTyped		If set the result data map will receive the value wrapped in an instance of <c>TypedValue</c>.
	# @param	bool bVarIsArray			Normally values in the result data map will be wrapped into an array only if multiple values occur with the
	#										same key. If this argument is set to <c>True</c> an array is created by default, even if it will receive only one
	#										value. This is useful if the values parsed will represent array data anyway.
	#
	def derive(self, assignToVar = None, assignToVarTyped = None, bVarIsArray = False):
		pass
	#

#













from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# A token pattern repeat pattern. The specified patteren must match at least once.
#
# @param		AbstractTokenPattern tokenPattern			(required) A token pattern that might be repeated multiple times.
#
class TokenPatternDelimLoop(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern tokenPattern			(required) A token pattern to match one or more times.
	#
	def __init__(self, tokenPattern:AbstractTokenPattern, delimPattern:AbstractTokenPattern):
		super().__init__()

		assert isinstance(tokenPattern, AbstractTokenPattern)
		self.__tokenPattern = tokenPattern

		assert isinstance(delimPattern, AbstractTokenPattern)
		self.__delimPattern = delimPattern
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		orgPos = pos

		(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
		if not bResult:
			return (False, 0)
		pos += n

		lastGoodPos = pos
		nCountLoops = 1

		while True:
			(bResult, n) = self.__delimPattern._tryMatch(tokens, pos, stack)
			if not bResult:
				break
			pos += n

			(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
			if not bResult:
				pos = lastGoodPos
				break
			pos += n

			nCountLoops += 1
			lastGoodPos = pos

		self._addTagsToStack(stack)
		return (True, pos - orgPos)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPatternDelimLoop(self.__tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
	#

#












class TokenizationError(Exception):

	def __init__(self, errMsg:str, lineNo:int, colNo:int, state:str = None):
		s = "Tokenization error encountered at " + str(lineNo) + ":" + str(colNo) + ": " + errMsg
		if state:
			s += " (Current state: " + repr(state) + ")"
		super().__init__(s)
	#

#



















from .MutableString import *
from .TextCanvas import *







#
# This object represents a table cell
#
class TextTableCell(object):

	def __init__(self, x, y, colSpan, rowSpan, textLines):
		assert isinstance(textLines, list)
		assert isinstance(colSpan, int)
		assert isinstance(rowSpan, int)
		assert x >= 0
		assert y >= 0
		assert colSpan >= 1
		assert rowSpan >= 1

		self.__x = x
		self.__y = y
		self.__textLines = textLines
		self.__maxTextWidth = 0
		for line in textLines:
			assert isinstance(line, str)
			if len(line) > self.__maxTextWidth:
				self.__maxTextWidth = len(line)
		self.__colSpan = colSpan
		self.__rowSpan = rowSpan
	#

	#
	# Tests if the specified position is covered by this cell. This method is useful especially for cells that span across multiple cells as this method
	# will return the correct results even in this case.
	#
	# @return		bool		Returns <c>True</c> or <c>False</c> wether the specified coordinates are covered by this cell or not.
	#
	def coversXY(self, x, y):
		return (x >= self.__x) and (x < self.__x + self.__colSpan) and (y >= self.__y) and (y < self.__y + self.__rowSpan)
	#

	def __repr__(self):
		s = ""
		for line in self.__textLines:
			s += "\n"
			s += line
		s = s[1:]
		if len(s) > 20:
			s = s[:20]
			s += "..."
		return "{ " + str(self.__x) + "," + str(self.__y) + " - " + str(self.__colSpan) + "," + str(self.__rowSpan) + " - " + repr(s) + " }"
	#

	def __str__(self):
		return self.__repr__()
	#

	#
	# The text lines stored in this cell
	#
	@property
	def textLines(self):
		return self.__textLines
	#

	#
	# The width the cell requires
	#
	@property
	def innerWidth(self):
		return self.__maxTextWidth
	#

	#
	# The height the cell requires
	#
	@property
	def innerHeight(self):
		return len(self.__textLines)
	#

	@property
	def x(self):
		return self.__x
	#

	@property
	def y(self):
		return self.__y
	#

	@property
	def rowSpan(self):
		return self.__rowSpan
	#

	@property
	def colSpan(self):
		return self.__colSpan
	#

#




#
# This class represents a table, which actually is a grid of cells. Text is organized in lines.
# After text has been added to the table you get it printed to an instance of <c>TextCanvas</c>
# which in turn can then be printed to STDOUT.
#
class TextTable(object):

	class __RowCreator(object):
		def __init__(self, table, rowNo):
			self.table = table
			self.rowNo = rowNo
			self.__pos = 0
		#
		def createCell(self, textLines, colSpan = 1, rowSpan = 1):
			while self.table.getCell(self.__pos, self.rowNo) != None:
				self.__pos += 1
			self.table.setCell(self.__pos, self.rowNo, textLines, colSpan, rowSpan)
			self.__pos += colSpan
			return self
		#
	#

	def __init__(self):
		self.__rows = []
		self.__countColumns = 0
		self.__headingRows = set()
	#

	def setHeadingRow(self, rowNo):
		self.__headingRows.add(rowNo)
	#

	def ensureSize(self, w, h):
		while len(self.__rows) < h:
			self.__rows.append([])
		self.__countColumns = max(w, self.__countColumns)
		for row in self.__rows:
			while len(row) < self.__countColumns:
				row.append(None)
	#

	def numberOfRows(self):
		return len(self.__rows)
	#

	def numberOfColumns(self):
		return self.__countColumns
	#

	def setCell(self, x, y, textLines, colSpan = 1, rowSpan = 1):
		if isinstance(textLines, str):
			textLines = [ textLines ]
		elif isinstance(textLines, list):
			pass
		else:
			raise Exception("Invalid data for text lines specified!")

		self.ensureSize(x + colSpan, y + rowSpan)
		self.__rows[y][x] = TextTableCell(x, y, colSpan, rowSpan, textLines)

		for ix in range(x, x + colSpan):
			for iy in range(y, y + rowSpan):
				if (ix != x) and (iy != y):
					self.__rows[y][x] = None
	#

	@staticmethod
	def __calcSpans(valueList, ofs, len):
		n = 0
		for i in range(ofs, ofs + len):
			n += valueList[ofs]
		return n
	#

	def _calcRowHeights(self):
		rowHeights = []
		numberOfRows = len(self.__rows)

		# get height of all cells that do not span across multiple cells

		multiRowCells = []
		for i in range(0, numberOfRows):
			rowHeights.append(0)
			row = self.__rows[i]
			for cell in row:
				if cell is None:
					continue
				if cell.rowSpan == 1:
					h = cell.innerHeight
					if h > rowHeights[i]:
						rowHeights[i] = h
				else:
					# collect the cells we need to deal with later
					multiRowCells.append(cell)

		# now make room for all cells spanning across multiple cells

		for cell in multiRowCells:
			currentHeightAvailable = TextTable.__calcSpans(rowHeights, cell.y, cell.rowSpan)	# get cell height
			currentHeightAvailable += cell.rowSpan - 1	# add intermediate borders
			# currentHeightAvailable is now the amount of height this cell can cover (without outer border)
			heightRequired = cell.innerHeight
			# heightRequired will contain the height the cell will need
			while currentHeightAvailable < heightRequired:
				# expand all heights by 1
				for i in range(0, cell.rowSpan):
					rowHeights[i] += 1
					currentHeightAvailable += 1

		return rowHeights
	#

	def _calcColumnWidths(self, extraHGapLeft, extraHGapRight):
		columnWidths = []
		numberOfRows = len(self.__rows)

		# get width of all cells that do not span across multiple cells

		#print("countColumns=" + str(self.__countColumns))
		for i in range(0, self.__countColumns):
			columnWidths.append(0)
		#print("columnWidths=" + str(columnWidths))
		#print("numberOfRows=" + str(numberOfRows))

		multiRowCells = []
		for i in range(0, numberOfRows):
			row = self.__rows[i]
			#print(str(i) + "\t" + str(row))
			j = -1
			for cell in row:
				j += 1
				if cell is None:
					continue
				if cell.colSpan == 1:
					w = cell.innerWidth + extraHGapLeft + extraHGapRight
					if w > columnWidths[j]:
						columnWidths[j] = w
				else:
					# collect the cells we need to deal with later
					multiRowCells.append(cell)

		# now make room for all cells spanning across multiple cells

		for cell in multiRowCells:
			currentWidthAvailable = TextTable.__calcSpans(columnWidths, cell.x, cell.colSpan)	# get cell width
			currentWidthAvailable += cell.colSpan - 1	# add intermediate borders
			# currentWidthAvailable is now the amount of height this cell can cover (without outer border)
			widthRequired = cell.innerWidth
			# widthRequired will contain the width the cell will need
			while currentWidthAvailable < widthRequired:
				# expand all widths by 1
				for i in range(0, cell.colSpan):
					columnWidths[i] += 1
					currentWidthAvailable += 1

		return columnWidths
	#

	def paintToTextCanvas(self, textCanvas = None, extraHGapLeft = 1, extraHGapRight = 1, bCompact = False):
		if textCanvas is None:
			textCanvas = TextCanvas()

		if bCompact:
			self.__paintToTextCanvasNoBorders(textCanvas, extraHGapLeft, extraHGapRight)
		else:
			self.__paintToTextCanvasWithBorders(textCanvas, extraHGapLeft, extraHGapRight)

		return textCanvas
	#

	def __paintToTextCanvasWithBorders(self, textCanvas, extraHGapLeft, extraHGapRight):
		columnWidths = self._calcColumnWidths(extraHGapLeft, extraHGapRight)
		xPositions = [ 0 ]
		for w in columnWidths:
			xPositions.append(xPositions[-1] + w + 1)

		rowHeights = self._calcRowHeights()
		yPositions = [ 0 ]
		for h in rowHeights:
			yPositions.append(yPositions[-1] + h + 1)

		#print("columnWidths=" + str(columnWidths))
		#print("xPositions=" + str(xPositions))
		#print("rowHeights=" + str(rowHeights))
		#print("yPositions=" + str(yPositions))

		textCanvas.ensureSize(xPositions[-1] + 1, yPositions[-1] + 1)

		for row in self.__rows:
			for cell in row:
				if cell is None:
					continue
				x1 = xPositions[cell.x]
				y1 = yPositions[cell.y]
				x2 = xPositions[cell.x + cell.colSpan]
				y2 = yPositions[cell.y + cell.rowSpan]
				textCanvas.drawRectangle(
					x1, y1, x2, y2,
					bDoubleBorderTop = (cell.y - 1) in self.__headingRows)
				textCanvas.drawTextLines(x1 + 1 + extraHGapLeft, y1 + 1, cell.textLines)

		return textCanvas
	#

	def __paintToTextCanvasNoBorders(self, textCanvas, extraHGapLeft, extraHGapRight):
		columnWidths = self._calcColumnWidths(extraHGapLeft, extraHGapRight)
		xPositions = [ 0 ]
		for w in columnWidths:
			xPositions.append(xPositions[-1] + w + 1)

		rowHeights = self._calcRowHeights()
		yPositions = [ 0 ]
		for h in rowHeights:
			yPositions.append(yPositions[-1] + h)

		#print("columnWidths=" + str(columnWidths))
		#print("xPositions=" + str(xPositions))
		#print("rowHeights=" + str(rowHeights))
		#print("yPositions=" + str(yPositions))

		textCanvas.ensureSize(xPositions[-1] + 1, yPositions[-1] + 1)

		for row in self.__rows:
			for cell in row:
				if cell is None:
					continue
				x1 = xPositions[cell.x]
				y1 = yPositions[cell.y]
				x2 = xPositions[cell.x + cell.colSpan]
				y2 = yPositions[cell.y + cell.rowSpan]
				textCanvas.drawTextLines(x1 + 1 + extraHGapLeft, y1 + 1, cell.textLines)

		return textCanvas
	#

	def createRow(self, bIsHeading = False):
		n = len(self.__rows)
		if bIsHeading:
			self.__headingRows.add(n)
		self.ensureSize(self.__countColumns, n + 1)
		return TextTable.__RowCreator(self, n)
	#

	#
	# Get the cell at position (x, y). This method is useful especially for cells that span across multiple cells as this method
	# will return the apropriate table cell even in this case.
	#
	def getCell(self, x, y):
		if (y >= len(self.__rows)) or (x >= self.__countColumns):
			return None
		iy = y
		while iy >= 0:
			ix = x
			while ix >= 0:
				cell = self.__rows[iy][ix]
				if cell != None:
					if cell.coversXY(x, y):
						return cell
				ix -= 1
			iy -= 1
		return None
	#

	def print(self, bCompact = False, printFunction = None):
		textCanvas = self.paintToTextCanvas(extraHGapLeft = 1, extraHGapRight = 1, bCompact = bCompact)
		textCanvas.print(printFunction = printFunction)
	#

	def dump(self, printFunction = None):
		if printFunction is None:
			printFunction = print
		else:
			assert callable(printFunction)

		printFunction("TextTable(")
		printFunction("\tcountColumns=" + str(self.__countColumns))
		printFunction("\tcountRows=" + str(len(self.__rows)))
		printFunction("\theadingRows=" + str(self.__headingRows))
		printFunction("\trows=[")
		i = 0
		for row in self.__rows:
			printFunction("\t\t" + str(i) + "\t" + str(row))
			i += 1
		printFunction("\t]")
		printFunction(")")
	#

#














#
# This class implements a data matrix for cells of unspecific type.
#
# TODO: Meanwhile we created an own package jk_datamatrix. Maybe merge this class into that package?
#
#
class DataMatrix(object):

	# ================================================================================================================================
	# ==== Constructor / Destructor
	# ================================================================================================================================

	def __init__(self, nCols:int, nRows:int):
		assert isinstance(nCols, int)
		assert nCols > 0
		assert isinstance(nRows, int)
		assert nRows > 0

		self.__rows = []
		for nRow in range(0, nRows):
			self.__rows.append([ None ] * nCols)

		self.__nRows = nRows
		self.__nCols = nCols
	#

	# ================================================================================================================================
	# ==== Properties
	# ================================================================================================================================

	#
	# Get the number of rows in this matrix.
	#
	# @return	int			Returns the number of rows.
	#
	@property
	def nRows(self) -> int:
		return self.__nRows
	#

	#
	# Get the number of colums in this matrix.
	#
	# @return	int			Returns the number of colums.
	#
	@property
	def nColumns(self) -> int:
		return self.__nCols
	#

	#
	# Get the size of this matrix.
	#
	# @return	tuple<int,int>			Returns the number of colums and rows (in this order).
	#
	@property
	def size(self) -> tuple:
		return self.__nCols, self.__nRows
	#

	# ================================================================================================================================
	# ==== Methods
	# ================================================================================================================================

	#
	# Append a matrix to the right of this matrix. For this to succeed both matrixes must have the same number of rows.
	#
	# @param	DataMatrix matrix		Another data matrix of the same height.
	#
	def appendMatrixToRight(self, matrix):
		assert isinstance(matrix, DataMatrix)
		assert matrix.__nRows == self.__nRows

		for rowIndex in range(self.__nRows):
			self.__rows[rowIndex].extend(matrix.__rows[rowIndex])
		self.__nCols += matrix.__nCols
	#

	#
	# Append a matrix to the bottom of this matrix. For this to succeed both matrixes must have the same number of columns.
	#
	# @param	DataMatrix matrix		Another data matrix of the same width.
	#
	def appendMatrixBelow(self, matrix):
		assert isinstance(matrix, DataMatrix)
		assert matrix.__nCols == self.__nCols

		self.__rows.extend([ list(r) for r in matrix.__rows])
		self.__nRows += matrix.__nRows
	#

	#
	# Remove a specific row.
	#
	# @param	int rowIndex		The index of the row to remove.
	#
	def removeRow(self, rowIndex:int):
		assert isinstance(rowIndex, int)
		if rowIndex < 0:
			rowIndex = self.__nRows + rowIndex

		del self.__rows[rowIndex]
		self.__nRows -= 1
	#

	#
	# Get the value of a specific cell.
	#
	def get(self, rowIndex:int, colIndex:int):
		if rowIndex < 0:
			rowIndex = self.__nRows + rowIndex
		if colIndex < 0:
			colIndex = self.__nCols + colIndex
		try:
			return self.__rows[rowIndex][colIndex]
		except IndexError as ee:
			self.dump()
			raise IndexError(str(ee) + " :: " + repr((rowIndex, colIndex)))
	#

	#
	# Set the value of a specific cell.
	#
	def set(self, rowIndex:int, colIndex:int, value):
		if rowIndex < 0:
			rowIndex = self.__nRows + rowIndex
		if colIndex < 0:
			colIndex = self.__nCols + colIndex
		try:
			self.__rows[rowIndex][colIndex] = value
		except IndexError as ee:
			self.dump()
			raise IndexError(str(ee) + " :: " + repr((rowIndex, colIndex)))
	#

	#
	# Add a row to the top of this matrix.
	#
	# @param	object content		(optional) An arbitrary object. The new row will be filled with references to the specified object.
	#
	def addRowAtTop(self, content = None):
		data = [ content for i in range(0, self.__nCols) ]
		self.__rows.insert(0, data)
		self.__nRows += 1
	#

	#
	# Add a row at the bottom of this matrix.
	#
	# @param	object content		(optional) An arbitrary object. The new row will be filled with references to the specified object.
	#
	def addRowAtBottom(self, content = None):
		data = [ content for i in range(0, self.__nCols) ]
		self.__rows.append(data)
		self.__nRows += 1
	#

	#
	# Add a column to the right of this matrix.
	#
	# @param	object content		(optional) An arbitrary object. The new column will be filled with references to the specified object.
	#
	def addColumnAtRight(self, content = None):
		for nRowIndex in range(0, self.__nRows):
			self.__rows[nRowIndex].append(content)
		self.__nCols += 1
	#

	#
	# Add a column to the left of this matrix.
	#
	# @param	object content		(optional) An arbitrary object. The new column will be filled with references to the specified object.
	#
	def addColumnAtLeft(self, content = None):
		for nRowIndex in range(0, self.__nRows):
			self.__rows[nRowIndex].insert(0, content)
		self.__nCols += 1
	#

	#
	# Get a copy of the data row by row as a two-dimensional array where the rows are returned as an immutable tuple.
	#
	def getRowTuple(self) -> tuple:
		return tuple([ tuple(row) for row in self.__rows ])
	#

	#
	# Get a copy of the data row by row as a two-dimensional array where the rows are returned as a list.
	#
	def getRowList(self) -> list:
		return [ list(row) for row in self.__rows ]
	#

	def dump(self):
		print(self.__nRows, "x", self.__nCols)
		for nRowIndex in range(0, self.__nRows):
			print("\t", nRowIndex, "::", len(self.__rows[nRowIndex]), str(self.__rows[nRowIndex]))
	#

#






import time
import threading
from typing import Union

import sortedcontainers





class _TimeLimitedCacheEntry(object):

	def __init__(self, identifier, deathTime:float, data):
		self.identifier = identifier
		self.deathTime = deathTime
		self.data = data
	#

#



def _getCEKey(item):
	return item.deathTime
#



#
# This class provides a cache that drops out items after a certain amount of time.
# Checks if data in the cache is obsolete is performed on every call of <c>put()</c>, <c>get()</c> and <c>remove()</c>.
# (If required by technical demands, this might be changed in the future.)
#
class TimeLimitedCache(object):

	def __init__(self, secondsToLive:Union[int,float]):
		assert isinstance(secondsToLive, (int, float))
		assert secondsToLive > 0

		self.__entriesByDeathTime = sortedcontainers.SortedListWithKey(key=_getCEKey)
		self.__entriesByID = dict()
		self.__secondsToLive = secondsToLive

		self.__lock = threading.Lock()
	#

	def __tryClean(self, tNow):
		while len(self.__entriesByDeathTime) > 0:
			existingEntry = self.__entriesByDeathTime[0]
			if existingEntry.deathTime > tNow:
				break
			else:
				self.__entriesByDeathTime.remove(existingEntry)
				del self.__entriesByID[existingEntry.identifier]
	#

	def put(self, identifier, item):
		assert isinstance(identifier, (str, int))

		tNow = time.time()

		newEntry = _TimeLimitedCacheEntry(identifier, tNow + self.__secondsToLive, item)

		self.__lock.acquire()

		existingEntry = self.__entriesByID.get(identifier, None)
		if existingEntry:
			self.__entriesByDeathTime.remove(existingEntry)
			del self.__entriesByID[identifier]

		self.__tryClean(tNow)

		self.__entriesByDeathTime.add(newEntry)
		self.__entriesByID[identifier] = newEntry

		self.__lock.release()
	#

	def get(self, identifier, bTouch:bool = True):
		assert isinstance(identifier, (str, int))
		assert isinstance(bTouch, bool)

		tNow = time.time()

		self.__lock.acquire()

		if bTouch:

			existingEntry = self.__entriesByID.get(identifier, None)
			if existingEntry:
				self.__entriesByDeathTime.remove(existingEntry)
				existingEntry.deathTime = tNow + self.__secondsToLive
				self.__entriesByDeathTime.add(existingEntry)

			self.__tryClean(tNow)

		else:

			self.__tryClean(tNow)
			existingEntry = self.__entriesByID.get(identifier, None)

		self.__lock.release()

		return None if existingEntry is None else existingEntry.data
	#

	def remove(self, identifier):
		assert isinstance(identifier, (str, int))

		self.__lock.acquire()

		existingEntry = self.__entriesByID.get(identifier, None)
		if existingEntry:
			self.__entriesByDeathTime.remove(existingEntry)
			del self.__entriesByID[identifier]

		self.__tryClean(time.time())

		self.__lock.release()
	#

#




























#!/usr/bin/python3



import weakref



class WeakRefObservableEvent(object):

	def __init__(self, name = None):
		self.__name = name
		self.__listeners = tuple()
	#

	@property
	def name(self):
		return self.__name
	#

	@property
	def listeners(self):
		return self.__listeners
	#

	def __len__(self):
		return len(self.__listeners)
	#

	def removeAllListeners(self):
		self.__listeners = tuple()
	#

	def __str__(self):
		if self.__name:
			ret = repr(self.__name)[1:][:-1] + "("
		else:
			ret = "Event("
		if len(self.__listeners) == 0:
			ret += "no listener)"
		elif len(self.__listeners) == 1:
			ret += "1 listener)"
		else:
			ret += str(len(self.__listeners)) + " listeners)"
		return ret
	#

	def __repr__(self):
		if self.__name:
			ret = repr(self.__name)[1:][:-1] + "("
		else:
			ret = "Event("
		if len(self.__listeners) == 0:
			ret += "no listener)"
		elif len(self.__listeners) == 1:
			ret += "1 listener)"
		else:
			ret += str(len(self.__listeners)) + " listeners)"
		return ret
	#

	def add(self, theCallable:callable):
		assert theCallable != None
		self.__listeners += (weakref.ref(theCallable),)
		return self
	#

	def __iadd__(self, theCallable:callable):
		assert theCallable != None
		self.__listeners += (weakref.ref(theCallable),)
		return self
	#

	def remove(self, theCallable:callable) -> bool:
		assert theCallable != None
		if theCallable in self.__listeners:
			n = self.__listeners.index(theCallable)
			self.__listeners = self.__listeners[:n] + self.__listeners[n + 1:]
			return True
		else:
			return False
	#

	def _removeAt(self, n:int):
		self.__listeners = self.__listeners[:n] + self.__listeners[n + 1:]
	#

	def __isub__(self, theCallable:callable):
		assert theCallable != None
		if theCallable in self.__listeners:
			n = self.__listeners.index(theCallable)
			self.__listeners = self.__listeners[:n] + self.__listeners[n + 1:]
		return self
	#

	def __call__(self, *argv, **kwargs):
		n = 0
		for listener in self.__listeners:
			o = listener()
			if o:
				o(*argv, **kwargs)
				n += 1
			else:
				self._removeAt(n)
	#

	def fire(self, *argv, **kwargs):
		n = 0
		for listener in self.__listeners:
			o = listener()
			if o:
				o(*argv, **kwargs)
				n += 1
			else:
				self._removeAt(n)
	#

#





import sys
import math




_PROGRESS_PARAMS_BYTES = [
	(1, "{:.2f}B"),
	(1, "{:.1f}B"),
	(1, "{:.0f}B"),
	(1000, "{:.2f}K"),
	(1000, "{:.1f}K"),
	(1000, "{:.0f}K"),
	(1000*1000, "{:.2f}M"),
	(1000*1000, "{:.1f}M"),
	(1000*1000, "{:.0f}M"),
	(1000*1000*1000, "{:.2f}G"),
	(1000*1000*1000, "{:.1f}G"),
	(1000*1000*1000, "{:.0f}G"),
]



def formatTime(seconds:float, withMilliseconds:bool = False):
	timeSeconds = int(seconds) % 60
	timeMinutes = int(seconds / 60)
	timeHours = int(timeMinutes / 60)
	timeMinutes = timeMinutes % 60
	if withMilliseconds:
		timeMillis = math.floor((seconds - int(seconds))*1000)
		return "{:02d}:{:02d}:{:02d}.{:03d}".format(timeHours, timeMinutes, timeSeconds, timeMillis)
	else:
		return "{:02d}:{:02d}:{:02d}".format(timeHours, timeMinutes, timeSeconds)
#



def formatBytes(nBytes:float):
	index = len(str(int(nBytes))) - 1
	if index >= len(_PROGRESS_PARAMS_BYTES):
		divider, formatStr = _PROGRESS_PARAMS_BYTES[-1]
	else:
		divider, formatStr = _PROGRESS_PARAMS_BYTES[index]
	return formatStr.format(nBytes / divider)
#



def formatBytesPerSecond(nBytesPerSecond:float):
	index = len(str(int(nBytesPerSecond))) - 1
	if index >= len(_PROGRESS_PARAMS_BYTES):
		divider, formatStr = _PROGRESS_PARAMS_BYTES[-1]
	else:
		divider, formatStr = _PROGRESS_PARAMS_BYTES[index]
	formatStr += "/s"
	return formatStr.format(nBytesPerSecond / divider)
#



#
# Displays or updates a console progress bar
#
# This is a (heavily) modified version of:
# https://stackoverflow.com/questions/3160699/python-progress-bar
#
# @param		float dTime					The amount of seconds already passed.
# @param		int currentCapacity			The amount of data already read/written.
# @param		int maxCapacity				The maximum number of data to read or write.
#
def showCapacityProgress(dTime:float, currentCapacity:int, maxCapacity:int):
	assert isinstance(dTime, float)
	assert isinstance(currentCapacity, int)
	assert isinstance(maxCapacity, int)
	assert maxCapacity > 0

	currentProgress = currentCapacity / maxCapacity
	if (dTime >= 1) and (currentCapacity > 0):
		totalTimeSeconds = dTime/currentProgress
		sTotalTime = formatTime(totalTimeSeconds)
		sUsedTime = formatTime(dTime)
		sDataRate = formatBytesPerSecond(currentCapacity / maxCapacity)
	else:
		sUsedTime = "??:??:??"
		sTotalTime = "??:??:??"
		sDataRate = "????"

	barLength = 40 # Modify this to change the length of the progress bar
	status = ""
	"""
	if currentProgress < 0:
		currentProgress = 0
		status = "Halt...\r\n"
	"""
	if currentProgress >= 1:
		currentProgress = 1
		status = "Done...\r\n"
	block = int(round(barLength*currentProgress))
	text = "\rProgress: [{0}] {1:.1f}% {2}   {3} ETA {4} @ {5}        ".format( "#"*block + "-"*(barLength-block), currentProgress*100, status, sUsedTime, sTotalTime, sDataRate)
	sys.stdout.write(text)
	sys.stdout.flush()
#





































class ChangedFlag(object):

	def __init__(self, value):
		assert isinstance(value, bool)
		self.__value = value
	#

	@property
	def value(self):
		return self.__value
	#

	@property
	def isChanged(self):
		return self.__value
	#

	def setChanged(self, value):
		assert isinstance(value, bool)
		self.__value = value
	#

	def __str__(self):
		if self.__value:
			return "changed"
		else:
			return "unchanged"
	#

	def __repr__(self):
		return self.__str__()
	#

	def __eq__(self, other):
		if isinstance(other, ChangedFlag):
			return other.__value == self.__value
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, ChangedFlag):
			return other.__value != self.__value
		else:
			return True
	#














import os
import re
import sys

_has_netifaces = False
try:
	import netifaces
	_has_netifaces = True
except:
	pass






#
# Retrieve all interface name, IP address, MAC address triplets from all active network adapters (excluding loopback and docker)
#
def getIPsEx():
	if _has_netifaces:
		ret = []

		for ifaceName in netifaces.interfaces():
			if ifaceName.startswith("docker"):
				continue
			if ifaceName.startswith("lo"):
				continue

			addrs = netifaces.ifaddresses(ifaceName)
			try:
				ifMac = addrs[netifaces.AF_LINK][0]["addr"]
				ipAddr = addrs[netifaces.AF_INET][0]["addr"]
				if ipAddr.startswith("127."):
					continue
				ret.append([ifaceName, ipAddr, ifMac])
			except IndexError as e:
				pass
			except KeyError as e:
				pass

		return ret

	else:
		raise Exception("Python module netiface module not available!")
#




#
# Retrieve all IP addresses from all active network adapters (excluding loopback and docker)
#
def getIPs():
	return [ x[1] for x in getIPsEx() ]
#






class LocalIPAddressDetector(object):

	def __init__(self, localIPAddressFilters:list):
		self.__localIPAddressRegExes = []
		for s in localIPAddressFilters:
			pos = s.index("*")
			if pos < 0:
				sRegex = "^" + s + "$"
			elif pos == 0:
				sRegex = "^.+" + s + "$"
			elif pos == len(s) - 1:
				sRegex = "^" + s + ".+$"
			else:
				raise Exception("Invalid network card pattern: " + repr(s))
			self.__localIPAddressRegExes.append(re.compile(sRegex))
	#

	def __call__(self):
		return self.getIPAddresses()
	#

	#
	# Retrieve all local IP addresses this machine might be contaced in a local network.
	#
	def getIPAddresses(self):
		if _has_netifaces:
			ret = []

			ipAddrTuples = []
			for ifaceName in netifaces.interfaces():
				addrs = netifaces.ifaddresses(ifaceName)
				try:
					#if_mac = addrs[netifaces.AF_LINK][0]["addr"]
					ipAddr = addrs[netifaces.AF_INET][0]["addr"]
					if ipAddr.startswith("127."):
						continue
					ipAddrTuples.append([ifaceName, ipAddr, False])
				except IndexError as e:
					pass
				except KeyError as e:
					pass

			for ipAddrRegEx in self.__localIPAddressRegExes:
				for r in ipAddrTuples:
					if r[2]:
						continue
					if ipAddrRegEx.match(r[0]):
						r[2] = True
						ret.append(r[1])

			return ret

		else:
			raise Exception("Python module netiface module not available!")
	#

#









import typing
import os
import subprocess
import signal

_bIsPOSIX = os.name == "posix"
if _bIsPOSIX:
	import pwd




#
# Kill a process.
#
# @param	int|dict process			A processes, either as <c>int</c> (= process ID) or as <c>dict</c>
#										as provided by <c>jk_sysinfo.get_ps()</c>.
# @param	AbstractLogger log			A logger that will receive log messages.
# @return	bool						Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
#										(In that case error log messages might have been generated, but no detailed error
#										information is returned.)
#
def killProcess(process:typing.Union[int,dict], log) -> bool:
	assert isinstance(process, (int,dict))

	# ----

	if process is None:
		if log:
			log.notice("No processes to kill.")
		return True

	ret = True

	try:
		if isinstance(process, int):
			pid = process
		else:
			assert isinstance(process, dict)
			pid = process["pid"]
			assert isinstance(pid, int)

		if log:
			log.notice("Killing: " + str(pid))
		os.kill(pid, signal.SIGTERM)
	except Exception as ee:
		if log:
			log.error(ee)
		ret = False

	return ret
#



#
# Kill a set of processes (in the order they are presented).
#
# @param	int[]|dict[] processes		A list of processes, either as <c>int[]</c> of process IDs or as <c>dict[]</c>
#										as provided by <c>jk_sysinfo.get_ps()</c>.
# @param	AbstractLogger log			A logger that will receive log messages.
# @return	bool						Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
#										(In that case error log messages might have been generated, but no detailed error
#										information is returned.)
#
def killProcesses(processes:typing.Union[list,tuple], bIgnoreNonExisting:bool, log=None) -> bool:
	assert isinstance(processes, (list, tuple))
	assert isinstance(bIgnoreNonExisting, bool)

	# ----

	if (processes is None) or (not processes):
		if log:
			log.notice("No processes to kill.")
		return True

	ret = True

	for process in processes:
		try:
			if isinstance(process, int):
				pid = process
			else:
				assert isinstance(process, dict)
				pid = process["pid"]
				assert isinstance(pid, int)

			if log:
				log.notice("Killing: " + str(pid))
			os.kill(pid, signal.SIGTERM)
		except ProcessLookupError as ee:
			if not bIgnoreNonExisting:
				if log:
					log.error(ee)
				ret = False
		except Exception as ee:
			if log:
				log.error(ee)
			ret = False

	return ret
#



def _demote(user_uid, user_gid):

	def fn():
		os.setgid(user_gid)
		os.setuid(user_uid)
	#

	return fn

#

if _bIsPOSIX:

	#
	# Run a process as other user and wait until it terminates. No input or output is processed.
	# Of course for this to work you should be user <c>root</c>.
	#
	# @param	str accountName			The name of the user account under which to execute the process.
	# @param	str filePath			The path of the file to execute. This can be a bash script.
	# @param	str[] arguments			(optional) A list of arguments for the program to execute.
	# @param	AbstractLogger log		A logger that will receive log messages.
	# @return	bool					Returns <c>true</c> on success. <c>false</c> is returned if something went wrong.
	#									(In that case error log messages might have been generated, but no detailed error
	#									information is returned.)
	#
	def runProcessAsOtherUser(accountName:str, filePath:str, args:typing.Union[list,tuple], log) -> bool:
		assert isinstance(accountName, str)
		assert accountName

		assert isinstance(filePath, str)
		assert filePath
		assert os.path.isfile(filePath)

		if args is None:
			args = []
		else:
			if isinstance(args, tuple):
				args = list(args)
			else:
				assert isinstance(args, list)

		# ----

		if log:
			log.notice("Running " + repr(filePath) + " as " + repr(accountName) + "...")

		scriptDirPath = os.path.dirname(filePath)

		pw_record = pwd.getpwnam(accountName)
		accountName = pw_record.pw_name
		user_home_dir = pw_record.pw_dir
		uid = pw_record.pw_uid
		gid = pw_record.pw_gid

		currentUserID = os.getuid()
		if currentUserID != 0:
			if uid != currentUserID:
				raise Exception("Must be root to run processes as other user!")

		env = os.environ.copy()
		env["HOME"     ]  = user_home_dir
		env["LOGNAME"  ]  = accountName
		env["PWD"      ]  = scriptDirPath
		env["USER"     ]  = accountName

		ret = True

		try:
			process = subprocess.Popen(
				[ filePath ] + args,
				preexec_fn = _demote(uid, gid),
				cwd =  scriptDirPath,
				env = env,
			)
			exitCode = process.wait()

			if exitCode != 0:
				if log:
					log.error("Script " + repr(filePath) + " terminated with exit code: " + str(exitCode))
				ret = False
			else:
				if log:
					log.notice("Done running " + repr(filePath) + ".")

		except Exception as ee:
			if log:
				log.error(ee)
			ret = False

		return ret
	#












from .MutableString import *




class TextCanvas(object):

	def __init__(self, width = 0):
		self.__rows = []
		self.__width = width
		self.__height = 0
	#

	def _newLine(self):
		self.__rows.append(MutableString(width = self.__width))
		self.__height += 1
	#

	def ensureSize(self, w, h):
		if w > self.__width:
			for i in range(0, self.__height):
				self.__rows[i].ensureWidth(w)
			self.__width = w
		while self.__height < h:
			self._newLine()
	#

	def drawRectangle(self, x1, y1, x2, y2, bDoubleBorderTop = False, bDoubleBorderBottom = False):
		assert isinstance(x1, int)
		assert isinstance(y1, int)
		assert isinstance(x2, int)
		assert isinstance(y2, int)

		self.ensureSize(x2 + 1, y2 + 1)
		for x in range(x1, x2):
			self.__rows[y1].setChar(x, "=" if bDoubleBorderTop else "-")
			self.__rows[y2].setChar(x, "=" if bDoubleBorderBottom else "-")
		for y in range(y1, y2):
			self.__rows[y].setChar(x1, "|")
			self.__rows[y].setChar(x2, "|")
		self.__rows[y1].setChar(x1, "+")
		self.__rows[y1].setChar(x2, "+")
		self.__rows[y2].setChar(x1, "+")
		self.__rows[y2].setChar(x2, "+")
	#

	def drawRectangleEx(self, x1, y1, x2, y2, charBorderTop = "-", charBorderRight = "|", charBorderBottom = "-", charBorderLeft = "|"):
		assert isinstance(x1, int)
		assert isinstance(y1, int)
		assert isinstance(x2, int)
		assert isinstance(y2, int)

		self.ensureSize(x2 + 1, y2 + 1)
		for x in range(x1, x2):
			self.__rows[y1].setChar(x, charBorderTop)
			self.__rows[y2].setChar(x, charBorderBottom)
		for y in range(y1, y2):
			self.__rows[y].setChar(x1, charBorderLeft)
			self.__rows[y].setChar(x2, charBorderRight)

		if (charBorderLeft != " ") or (charBorderRight != " "):
			self.__rows[y1].setChar(x1, "+")	# top left
		if (charBorderRight != " ") or (charBorderRight != " "):
			self.__rows[y1].setChar(x2, "+")	# top right
		if (charBorderLeft != " ") or (charBorderBottom != " "):
			self.__rows[y2].setChar(x1, "+")	# bottom left
		if (charBorderRight != " ") or (charBorderBottom != " "):
			self.__rows[y2].setChar(x2, "+")	# bottom right
	#

	def drawTextLines(self, x, y, textLines):
		assert isinstance(x, int)
		assert isinstance(y, int)
		assert isinstance(textLines, list)

		maxW = 0
		for textLine in textLines:
			if len(textLine) > maxW:
				maxW = len(textLine)
		self.ensureSize(x + maxW + 1, y + len(textLines) + 1)

		for textLine in textLines:
			self.__rows[y].setText(x, textLine)
			y += 1
	#

	def drawText(self, x, y, textLine):
		assert isinstance(x, int)
		assert isinstance(y, int)
		assert isinstance(textLine, str)

		self.ensureSize(x + len(textLine) + 1, y + 1)

		self.__rows[y].setText(x, textLine)
	#

	def print(self, printFunction = None):
		if printFunction is None:
			printFunction = print
		else:
			assert callable(printFunction)

		for line in self.__rows:
			printFunction(str(line))
	#

	def toTextLines(self, prefix = None, bRightStrip = True):
		if prefix is None:
			prefix = ""

		ret = []
		for line in self.__rows:
			s = prefix + str(line)
			if bRightStrip:
				s = s.rstrip()
			ret.append(s)
		return ret
	#

	def toString(self, prefix = None, bRightStrip = True):
		if prefix is None:
			prefix = ""

		ret = ""
		for line in self.__rows:
			s = prefix + str(line)
			if bRightStrip:
				s = s.rstrip()
			ret += s + "\n"
		return ret
	#

	def __str__(self):
		return self.toString()
	#

	def __repr__(self):
		return "TextCanvas(width=" + str(self.__width) + ", height=" + str(self.__height) + ")"
	#

#

















__version__ = "0.2022.2.6"


from .Token import Token
from .RegExBasedTableTokenizer import RegExBasedTableTokenizer, RegExBasedTokenizingTable
from .RegExBasedTokenizer import RegExBasedTokenizer

from .TokenizationError import TokenizationError

from .AbstractTokenPattern import AbstractTokenPattern
from .TokenPattern import TokenPattern
from .TokenPatternAlternatives import TokenPatternAlternatives
from .TokenPatternRepeat import TokenPatternRepeat
from .TokenPatternRepeatUntilNot import TokenPatternRepeatUntilNot
from .TokenPatternSequence import TokenPatternSequence
from .TokenPatternDelimLoop import TokenPatternDelimLoop





from .RGB import RGB




class RGBSpectrum(object):

	def __init__(self, rgbValues:list):
		self.__values = rgbValues
	#

	def length(self) -> int:
		return len(self.__values)
	#

	def __len__(self) -> int:
		return len(self.__values)
	#

	def getRange(self, nStart:int, nEnd:int):
		return self.__values[nStart:nEnd]
	#

	def getE(self, n:int):
		return self.__values[n]
	#

	def getN(self, n:int):
		if (n < 0) or (n >= len(self.__values)):
			return None
		return self.__values[n]
	#

	def get(self, n:int):
		if n < 0:
			n = 0
		if n >= len(self.__values):
			n = len(self.__values) - 1
		return self.__values[n]
	#

	def __iter__(self):
		return self.__values.__iter__()
	#

#









import inspect




def isStaticMethod(theClass, methodName):
	try:
		value = getattr(theClass, methodName)
	except:
		# Such a method does not exist
		return False

	assert getattr(theClass, methodName) == value

	for cls in inspect.getmro(theClass):
		if inspect.isroutine(value):
			if methodName in cls.__dict__:
				binded_value = cls.__dict__[methodName]
				if isinstance(binded_value, staticmethod):
					return True
	return False
#









import datetime
import re





#
# This class represents a date.
#
class D(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self):
		self._dt = None
	#

	################################################################################################################################
	## Properties
	################################################################################################################################

	#
	# The time stamp in seconds
	#
	@property
	def ts(self) -> int:
		return int(self._dt.timestamp())
	#

	#
	# The time stamp in nanoseconds.
	#
	def tsNano(self) -> int:
		return int(self._dt.timestamp() * 1000000)
	#

	@property
	def year(self) -> int:
		return self._dt.year
	#

	#
	# The month of this year, starting with one.
	#
	@property
	def month(self) -> int:
		return self._dt.month
	#

	#
	# The month of this year, starting with zero.
	#
	@property
	def month0(self) -> int:
		return self._dt.month - 1
	#

	#
	# The day of this month, starting with one.
	#
	@property
	def day(self) -> int:
		return self._dt.day
	#

	#
	# The day of this month, starting with zero.
	#
	@property
	def day0(self) -> int:
		return self._dt.day - 1
	#

	#
	# The week day: Monday = 1, Tuesday = 2, ..., Sunday = 7
	#
	@property
	def dayOfWeek(self) -> int:
		return self._dt.weekday() + 1
	#

	#
	# The week day: Monday = 0, Tuesday = 1, ..., Sunday = 6
	#
	@property
	def dayOfWeek0(self) -> int:
		return self._dt.weekday()
	#

	#
	# The week day: Monday = 1, Tuesday = 2, ..., Sunday = 7
	#
	@property
	def weekday(self) -> int:
		return self._dt.weekday() + 1
	#

	#
	# The week day: Monday = 0, Tuesday = 1, ..., Sunday = 6
	#
	@property
	def weekday0(self) -> int:
		return self._dt.weekday() + 1
	#

	#
	# This date in 8 digit year-month-day representaion.
	#
	@property
	def yearMonthDay(self) -> int:
		return self._dt.year * 10000 + self._dt.month * 100 + self._dt.day
	#

	#
	# This date in 6 digit year-month representaion.
	#
	@property
	def yearMonth(self) -> int:
		return self._dt.year * 100 + self._dt.month
	#

	"""
	#
	# Calculates the week number (using ISO week number algorithm, where Jan 1st is in week 1)
	#
	@property
	def weekNo(self):
		d = datetime.datetime(self.year, 1, 1)

		# the week day of the first day in this year:
		# 1 = monday, 2 = tuesday, ..., 7 = sunday
		firstJanWeekDay = ((d.weekday() + 1) + 7) % 7
		correction = 0 if firstJanWeekDay == 0 else (7 - (- firstJanWeekDay + 1))
		td = self._dt - d
		return (td.days + correction) // 7
	#
	"""

	#
	# Official week number (where the week, that has the majority of days in this year, is week 1)
	#
	@property
	def weekNo(self):
		return self._dt.isocalendar()[1]
	#

	################################################################################################################################
	## Special Methods
	################################################################################################################################

	def __str__(self):
		return "{:04d}-{:02d}-{:02d}".format(self._dt.year, self._dt.month, self._dt.day)
	#

	def __repr__(self):
		return "{:04d}-{:02d}-{:02d}".format(self._dt.year, self._dt.month, self._dt.day)
	#

	def __hash__(self):
		return (self._dt.year * 10000 + self._dt.month * 100 + self._dt.day).__hash__()
	#

	################################################################################################################################
	## Operator Methods
	################################################################################################################################

	def __eq__(self, other):
		if isinstance(other, D):
			return self._dt == other._dt
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, D):
			return self._dt != other._dt
		else:
			return True
	#

	def __ge__(self, other):
		assert isinstance(other, D)
		return self._dt >= other._dt
	#

	def __gt__(self, other):
		assert isinstance(other, D)
		return self._dt > other._dt
	#

	def __le__(self, other):
		assert isinstance(other, D)
		return self._dt <= other._dt
	#

	def __lt__(self, other):
		assert isinstance(other, D)
		return self._dt < other._dt
	#

	def __add__(self, other):
		if isinstance(other, int):
			return D.createFrom(dt=int(self._dt.timestamp() + 24*60*60*other))
		else:
			raise Exception(str(type(other)))
	#

	def __iadd__(self, other):
		if isinstance(other, int):
			self._dt = datetime.datetime.fromtimestamp(self._dt.timestamp() + 24*60*60*other)
			return self
		else:
			raise Exception(str(type(other)))
	#

	def __sub__(self, other):
		if isinstance(other, int):
			return D.createFrom(dt=int(self._dt.timestamp() - 24*60*60*other))
		elif isinstance(other, D):
			delta = self._dt.timestamp() - other._dt.timestamp()
			return int(delta / (24*60*60))
		else:
			raise Exception(str(type(other)))
	#

	def __isub__(self, other):
		if isinstance(other, int):
			self._dt = datetime.datetime.fromtimestamp(self._dt.timestamp() - 24*60*60*other)
			return self
		else:
			raise Exception(str(type(other)))
	#

	def __int__(self):
		return int(self._dt.timestamp())
	#

	def __float__(self):
		return self._dt.timestamp()
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def toISOStr(self):
		return "{:04d}-{:02d}-{:02d}".format(self._dt.year, self._dt.month, self._dt.day)
	#

	def toEUStr(self):
		return "{:02d}.{:02d}.{:04d}".format(self._dt.day, self._dt.month, self._dt.year)
	#

	def toJSON(self) -> dict:
		return {
			"year": self._dt.year,
			"month": self._dt.month,
			"day": self._dt.day,
			"ts": int(self._dt.timestamp()),
			"yearMonth": self.yearMonth,
			"yearMonthDay": self.yearMonthDay,
			"isostr": str(self),
		}
	#

	#
	# Creates a D object representing the first of January of this year
	#
	# @return		D		A new date object
	#
	def startOfYear(self):
		ret = D()
		ret._dt = datetime.datetime(self.year, 1, 1)
		return ret
	#

	#
	# Creates a D object representing the 31st of December of this year
	#
	# @return		D		A new date object
	#
	def endOfYear(self):
		ret = D()
		ret._dt = datetime.datetime(self.year, 12, 31)
		return ret
	#

	#
	# @return		D		A new date object
	#
	def startOfMonth(self):
		ret = D()
		ret._dt = datetime.datetime(self.year, self.month, 1)
		return ret
	#

	#
	# @return		D		A new date object
	#
	def endOfMonth(self):
		y = self.year
		m = self.month + 1
		if m > 12:
			y += 1
			m -= 12
		dt = datetime.datetime(y, m, 1)
		dt = dt - datetime.timedelta(days=1)

		ret = D()
		ret._dt = dt
		return ret
	#

	#
	# @return		D		A new date object
	#
	def nextYear(self):
		ret = D()
		ret._dt = datetime.datetime(self.year + 1, self.month, self.day)
		return ret
	#

	#
	# @return		D		A new date object
	#
	def previousYear(self):
		if self.year <= 1:
			raise Exception()
		ret = D()
		ret._dt = datetime.datetime(self.year - 1, self.month, self.day)
		return ret
	#

	#
	# @return		D		A new date object
	#
	def nextMonth(self):
		y = self.year
		m = self.month
		d = self.day

		if m == 12:
			m = 1
			y += 1
		else:
			m += 1

		ret = D()
		while True:
			try:
				ret._dt = datetime.datetime(y, m, d)
				break
			except ValueError as ee:
				d -= 1
		return ret
	#

	#
	# @return		D		A new date object
	#
	def previousMonth(self):
		y = self.year
		m = self.month
		d = self.day

		if m == 1:
			m = 12
			y -= 1
		else:
			m -= 1

		ret = D()
		while True:
			try:
				ret._dt = datetime.datetime(y, m, d)
				break
			except ValueError as ee:
				d -= 1
		return ret
	#

	#
	# @return		D		A new date object
	#
	def startOfWeek(self):
		wd = self._dt.weekday()
		dt = self._dt.timestamp() - 24*60*60*wd
		d = D()
		d._dt = datetime.datetime.fromtimestamp(dt)
		return d
	#

	#
	# @return		D		A new date object
	#
	def endOfWeek(self):
		wd = self._dt.weekday()
		dt = self._dt.timestamp() - 24*60*60*wd + 24*60*60*6
		d = D()
		d._dt = datetime.datetime.fromtimestamp(dt)
		return d
	#

	#
	# @return		D		A new date object
	#
	def nextWeek(self):
		wd = self._dt.weekday()
		dt = self._dt.timestamp() - 24*60*60*wd
		d = D()
		d._dt = datetime.datetime.fromtimestamp(dt + 7 *24*60*60)
		return d
	#

	#
	# @return		D		A new date object
	#
	def previousWeek(self):
		wd = self._dt.weekday()
		dt = self._dt.timestamp() - 24*60*60*wd
		d = D()
		d._dt = datetime.datetime.fromtimestamp(dt - 7 *24*60*60)
		return d
	#

	#
	# @return		D		A new date object
	#
	def nextDay(self):
		d = D()
		d._dt = datetime.datetime.fromtimestamp(self._dt.timestamp() + 24*60*60)
		return d
	#

	#
	# @return		D		A new date object
	#
	def previousDay(self):
		d = D()
		d._dt = datetime.datetime.fromtimestamp(self._dt.timestamp() - 24*60*60)
		return d
	#

	#
	# @return		D		A new date object
	#
	def addMonths(self, m:int):
		assert isinstance(m, int)

		yy = self.year
		mm = self.month
		dd = self.day

		if m > 0:
			mm += m % 12
			yy += m // 12
		else:
			m = -m
			mm -= m % 12
			yy -= m // 12
			if mm < 1:
				yy -= 1
				mm += 12
		
		return D.createFrom(yearMonthDayTuple=(yy, mm, dd))
	#

	#
	# @return		D		A new date object
	#
	def subtractMonths(self, m:int):
		assert isinstance(m, int)

		return self.addMonths(-m)
	#

	#
	# @return		D		A new date object
	#
	def addYears(self, y:int):
		assert isinstance(y, int)

		yy = self.year + y
		mm = self.month
		dd = self.day

		return D.createFrom(yearMonthDayTuple=(yy, mm, dd))
	#

	#
	# @return		D		A new date object
	#
	def subtractYears(self, y:int):
		assert isinstance(y, int)

		return self.addYears(-y)
	#

	#
	# @return		D		A new date object
	#
	def clone(self):
		ret = D()
		ret._dt = self._dt
		return ret
	#

	#
	# @return	D[]		Returns a list of date objects containing all seven days of this week.
	#
	def generateWeekDates(self) -> list:
		d = self.startOfWeek()
		ret = [ d ]
		for i in range(0, 6):
			d = d.nextDay()
			ret.append(d)
		return ret
	#

	################################################################################################################################
	## Static Methods
	################################################################################################################################

	#
	# Returns a date object of today.
	# This is the same as "today()`.
	#
	@staticmethod
	def now():
		ret = D()
		dt = datetime.datetime.now()
		y = dt.year
		m = dt.month
		d = dt.day
		ret._dt = datetime.datetime(y, m, d)
		return ret
	#

	#
	# Returns a date object of today.
	# This is the same as "now()`.
	#
	@staticmethod
	def today():
		ret = D()
		dt = datetime.datetime.now()
		y = dt.year
		m = dt.month
		d = dt.day
		ret._dt = datetime.datetime(y, m, d)
		return ret
	#

	@staticmethod
	def tryParseFromStr(s:str):
		if isinstance(s, str):
			s = s.strip()
			for dt, year, month, day in D.__tryParse(s):
				if dt is None:
					if year is None:
						d0 = D.now()
						year = d0._dt.year
						try:
							d1 = D.createFrom(yearMonthDayTuple=(year, month, day))
							if d1 < d0:
								d1 = D.createFrom(yearMonthDayTuple=(year + 1, month, day))
							return d1
						except:
							pass
					else:
						try:
							return D.createFrom(yearMonthDayTuple=(year, month, day))
						except:
							pass
				else:
					try:
						return D.createFrom(dt=dt)
					except:
						pass
		return None
	#

	@staticmethod
	def parseFromStr(s:str):
		d = D.tryParseFromStr(s)
		if d is None:
			raise Exception("Failed to parse: " + repr(s))
		return d
	#

	@staticmethod
	def __tryParse(s:str):
		for patternType, pattern in [
			("ymd", r"(?P<year>\d\d\d\d)-(?P<month>\d\d)-(?P<day>\d\d)"),
			("ymd", r"(?P<day>\d\d?)\.(?P<month>\d\d?)\.(?P<year>\d\d\d\d)"),
			("ymd", r"(?P<month>\d\d?)/(?P<day>\d\d?)/(?P<year>\d\d\d\d)"),
			("md", r"(?P<day>\d\d?)\.(?P<month>\d\d?)\."),
			("ts", r"(?P<ts>-?\d+)"),
		]:
			m = re.match("^" + pattern + "$", s)
			if m:
				if patternType == "ts":
					try:
						yield (int(s), None, None, None)
					except:
						pass
				elif patternType == "ymd":
					year = int(m.group("year"))
					month = int(m.group("month"))
					day = int(m.group("day"))
					if (1 <= month <= 12) and (1 <= day <= 31) and (100 <= year <= 2999):
						yield (None, year, month, day)
				elif patternType == "md":
					month = int(m.group("month"))
					day = int(m.group("day"))
					if (1 <= month <= 12) and (1 <= day <= 31):
						yield (None, None, month, day)
				else:
					raise Exception()
	#

	@staticmethod
	def createFrom(something = None, dt:int = None, yearMonthDay:int = None, yearMonth:int = None, yearMonthDayTuple:list = None):
		ret = D()

		if dt is not None:
			if isinstance(dt, datetime.datetime):
				ret._dt = D.__createFrom_timeStamp(dt)
			elif isinstance(dt, (int, float)):
				ret._dt = D.__createFrom_timeStamp(dt)
			else:
				raise Exception()
		elif yearMonthDay is not None:
			ret._dt = D.__createFrom_yearMonthDay(yearMonthDay)
		elif yearMonth is not None:
			ret._dt = D.__createFrom_yearMonth(yearMonth)
		elif yearMonthDayTuple is not None:
			ret._dt = D.__createFrom_yearMonthDayTuple(yearMonthDayTuple)
		elif something is not None:
			if isinstance(something, datetime.datetime):
				ret._dt = something
			elif isinstance(something, float):
				ret._dt = D.__createFrom_timeStamp(something)
			elif isinstance(something, int):
				if 10000000 < something < 29999999:
					ret._dt = D.__createFrom_yearMonthDay(something)
				elif 100000 < something < 299999:
					ret._dt = D.__createFrom_yearMonth(something)
				elif something > 999999999:
					ret._dt = D.__createFrom_timeStamp(something)
				else:
					raise Exception()
			elif isinstance(something, str):
				ret = D.parseFromStr(something)
			elif isinstance(something, (tuple, list)):
				ret._dt = D.__createFrom_yearMonthDayTuple(something)
			else:
				raise Exception()
		else:
			raise Exception()

		return ret
	#

	@staticmethod
	def __createFrom_timeStamp(timeStamp:float) -> datetime.datetime:
		dt = datetime.datetime.fromtimestamp(timeStamp)
		year = dt.year
		month = dt.month
		day = dt.day
		return datetime.datetime(year, month, day)
	#

	@staticmethod
	def __createFrom_yearMonth(yearMonth:int) -> datetime.datetime:
		year = yearMonth // 100
		month = yearMonth % 100
		day = 1
		return datetime.datetime(year, month, day)
	#

	@staticmethod
	def __createFrom_yearMonthDay(yearMonthDay:int) -> datetime.datetime:
		year = yearMonthDay // 10000
		month = (yearMonthDay // 100) % 100
		day = yearMonthDay % 100
		return datetime.datetime(year, month, day)
	#

	@staticmethod
	def __createFrom_yearMonthDayTuple(yearMonthDayTuple) -> datetime.datetime:
		assert isinstance(yearMonthDayTuple, (list, tuple))
		assert len(yearMonthDayTuple) == 3
		assert isinstance(yearMonthDayTuple[0], int)
		assert isinstance(yearMonthDayTuple[1], int)
		assert isinstance(yearMonthDayTuple[2], int)
		return datetime.datetime(yearMonthDayTuple[0], yearMonthDayTuple[1], yearMonthDayTuple[2])
	#

#





















import collections



#
# This defines a token data structure. This structure has the following fields:
#
# @attribute		str type		A type identifier that specifies the type of the token
# @attribute		obj value		The value of the token (can be of any data type)
# @attribute		int lineNo		The line number (counted from 1)
# @attribute		int colNo		The character number within the line (counted from 1)
#
Token = collections.namedtuple("Token", ["type", "value", "lineNo", "colNo", "length"])














class TabularWriterSTDOUT(object):

	def __init__(self, columns:list, colWidth:int):
		self.__columns = columns
		self.__colWidth = colWidth

		s = ""
		self.__proto = []
		for i in range(0, len(columns)):
			self.__proto.append("-")
			n = i * colWidth
			while len(s) < n:
				s += " "
			s += columns[i]
		print(s)

		n = len(columns) * colWidth
		print("=" * n)

		self.__separator = "- " * (n // 2)
	#

	def print(self, column:str, text:str):
		outList = list(self.__proto)
		outList[self.__columns.index(column)] = text

		s = ""
		for item in outList:
			while len(item) < self.__colWidth:
				item += " "
			s += item
		print(s)

		print(self.__separator)
	#

	def close(self):
		pass
	#

#











import typing



def allItemsHaveEqualLength(listOfSequences) -> bool:
	it = iter(listOfSequences)
	firstLength = len(next(it))
	if not all( len(item) == firstLength for item in it ):
		return False
	return True
#

def allItemsHaveEqualLengthE(listOfSequences) -> int:
	it = iter(listOfSequences)
	firstLength = len(next(it))
	if not all( len(item) == firstLength for item in it ):
		raise Exception("Sequence contains components of different length!")
	return firstLength
#
















#
# This class implements a stack.
#
# @TODO		Some optimizations might be needed.
#
class Stack(object):

	class _Mark(object):

		def __init__(self, stack, pos):
			self.__stack = stack
			self.__pos = pos
		#

		def resetToMark(self):
			self.__stack._data = self.__stack._data[0:self.__pos]
		#

	#

	def __init__(self):
		self._data = []
	#

	def __iter__(self):
		for v in self._data:
			yield v
	#

	def __len__(self):
		return len(self._data)
	#

	def __str__(self):
		ret = "Stack["
		bAddComma = False
		for item in self._data:
			if bAddComma:
				ret += ", "
			else:
				bAddComma = True
			ret += str(item)
		return ret + "]"
	#

	def __repr__(self):
		return self.__str__()
	#

	def push(self, item):
		self._data.append(item)
	#

	def pop(self):
		if len(self._data) == 0:
			raise Exception("No item to pop!")
		ret = self._data[len(self._data) - 1]
		self._data = self._data[0:-1]
		return ret
	#

	def mark(self):
		return Stack._Mark(self, len(self._data))
	#

#














__version__ = "0.2022.2.6"


from .D import D
from .T1 import T1
from .T5 import T5
from .TApprox import TApprox

from .dateiterator import dateRange, dateRangeI
from .timeiterator5 import timeRange5






from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# A token pattern repeat pattern. The specified patteren must match at least once.
#
# @param		AbstractTokenPattern tokenPattern			(required) A token pattern that might be repeated multiple times.
#
class TokenPatternRepeatUntilNot(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern tokenPattern			(required) A token pattern to match one or more times.
	#
	def __init__(self, tokenPattern:AbstractTokenPattern, assignTokensToVar:str = None):
		super().__init__()

		assert isinstance(tokenPattern, AbstractTokenPattern)
		self.__tokenPattern = tokenPattern

		assert isinstance(assignTokensToVar, (type(None), str))
		self.__assignTokensToVar = assignTokensToVar
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		orgPos = pos

		while True:
			if pos == len(tokens):
				return (False, 0)

			(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if bResult:
				break

			pos += 1

		if self.__assignTokensToVar is not None:
			stack.push(("d", tokens[orgPos:pos], self.__assignTokensToVar, False))
		self._addTagsToStack(stack)
		return (True, pos - orgPos)
	#

#












import time
import threading



from .ObservableEvent import ObservableEvent



#
# TODO: Make the timing more accurate by using a more precise sleep time!
#
class Timer(object):

	def __init__(self, delaySeconds:int):
		self.__bKeepRunning = True
		self.__delaySeconds = delaySeconds
		self.__onTimerEvent = ObservableEvent("timerTick")
		self.__onErrorEvent = ObservableEvent("timerError", bCatchExceptions=True)

		self.__thread = threading.Thread(target=self.__runTimer, daemon=True)
		self.__thread.start()
	#

	def __del__(self):
		self.__bKeepRunning = False
	#

	@property
	def onTimer(self):
		return self.__onTimerEvent
	#

	@property
	def onError(self):
		return self.__onErrorEvent
	#

	def __runTimer(self):
		while self.__bKeepRunning:
			time.sleep(self.__delaySeconds)
			try:
				self.__onTimerEvent(self)
			except Exception as ee:
				self.__onErrorEvent(ee)

		self.__thread = None
	#

	def terminate(self):
		self.__bKeepRunning = False
	#

#















import os
import sys
import socket
import time
import json
import codecs
import re

_has_netifaces = False
try:
	import netifaces
	_has_netifaces = True
except:
	pass

import jk_simpleexec

from .dtutils import *
from .Cache import Cache






class ArpRecord(object):

	def __init__(self, ipAddress, flag, macAddr, interface):
		self.ipAddress = ipAddress
		#self.flag = flag
		self.macAddress = macAddr
		self.interface = interface
	#

	def dump(self, out_stream = sys.stdout, prefix = ""):
		out_stream.write(prefix + self.__repr__() + "\n")
	#

	def __repr__(self):
		return "ARP(" + self.ipAddress + ", " + self.macAddress + ", " + self.interface + ")"
	#

	def __str__(self):
		return "ARP(" + self.ipAddress + ", " + self.macAddress + ", " + self.interface + ")"
	#

#



class cachedarp(object):

	__cache = Cache(60)

	@staticmethod
	def arpResolveByIP(ipAddress):
		r = cachedarp.__cache.get(ipAddress)
		if r != None:
			return r[1]
		mac = arp.arpResolveByIP(ipAddress)
		cachedarp.__cache.put(ipAddress, (True, mac))
		return mac
	#

#




class arp(object):

	__reReply1 = re.compile("^ARPING [0-9\.]+ from [0-9\.]+ ([0-9a-zA-Z\.]+)$")
	__reReply2 = re.compile("^Unicast reply from [0-9\.]+ \[([0-9a-fA-F:]+)\] ")


	#
	# Get the MAC address for the specified IP address.
	# Please note that data is taken from the operating system cache if possible.
	# If no record exists in the cache an ARP query is performed. During this query the remote host has a single second to reply,
	# so this method will take at most one second to execute.
	#
	# @param		str ipAddress		The IP address to get the MAC address for
	# @return		ArpRecord			Either returns a valid <c>ArpRecord</c> object or <c>None</c>.
	#
	@staticmethod
	def arpResolveByIP(ipAddress):
		assert isinstance(ipAddress, str)

		for arpEntry in arp.systemArpEntries():
			if arpEntry.ipAddress == ipAddress:
				return arpEntry
		for arpEntry in arp.arpEntriesOfLocalInterfaces():
			if arpEntry.ipAddress == ipAddress:
				return arpEntry
		arpEntry = arp.arpQuery(ipAddress)
		return arpEntry
	#



	#
	# Get the IP address for the specified MAC address.
	# Please note that data is taken from the operating system cache only. If no data is stored in that cache this method will return <c>None</c>.
	#
	# @param		str macAddress		The MAC address to get the IP address for
	# @return		ArpRecord			If no data is in the cache <c>None</c> is returned. Otherwise a valid <c>ArpRecord</c> object is returned.
	#
	@staticmethod
	def arpResolveByMAC(macAddress):
		assert isinstance(macAddress, str)

		for arpEntry in arp.systemArpEntries():
			if arpEntry.macAddress == macAddress:
				return arpEntry
		return None
	#



	#
	# Perform an single ARP query.
	# Please note that invoking this method will cause network traffic. A remote host is queried. As this host has a single second to reply
	# this method will take at most one second to execute.
	# Please also note that remote hosts that are not connected well might exceed the timeout and therefor will not be listed by this method.
	#
	# @param		str ipAddress		The IP address to get the MAC address for
	# @return		ArpRecord			Returns an <c>ArpRecord</c> object or <c>None</c>.
	#
	@staticmethod
	def arpQuery(ipAddress):
		assert isinstance(ipAddress, str)

		result = jk_simpleexec.invokeCmd("/usr/bin/arping", [ "-w", "0.5", "-fb", "-c", "1", ipAddress ])
		if len(result.stdErrLines) > 0:
			raise Exception("arping failed: " + result.commandPath + " " + str(result.commandArguments))
		lines = result.stdOutLines
		result1 = arp.__reReply1.match(lines[0])
		if result1:
			result2 = arp.__reReply2.match(lines[1])
			if result2:
				return ArpRecord(ipAddress, -1, result2.group(1), result1.group(1))
			return None
		else:
			raise Exception("arping failed: Unexpected output: " + lines[0])
	#



	#
	# Queries the operating system ARP cache to list all known ARP records.
	# Please note that if an ARP record is missing this does not mean that this host is not active.
	# This only means the operating system just had no recent contact with this system.
	# Please also note that local IP and MAC addresses are not listed in this cache. User <c>arpEntriesOfLocalInterfaces()</c>
	# to list pseudo ARP entries for these network interfaces.
	#
	# @return		ArpRecord[]			Returns (a possibly empty list) of ARP records.
	#
	@staticmethod
	def systemArpEntries():
		with codecs.open("/proc/net/arp", "r") as f:
			lines = f.readlines()
			lines = lines[1:]
		ret = []
		for line in lines:
			ipAddress, hexHWType, hexFlag, macAddr, mask, interfaceDevice = re.split(" +", line[:-1])
			flag = int(hexFlag, 0)
			if flag in [2, 4, 6, 8]:
				ret.append(ArpRecord(ipAddress, flag, macAddr, interfaceDevice))
		return ret
	#



	#
	# Returns a list of MACs that are used locally by this host.
	# Please note that this method addresses only the local interfaces as they these MAC addresses are not maintained by the
	# operating system ARP cache.
	#
	# @return		ArpRecord[]			Returns a (possibly empty) list of <c>ArpRecord</c> objects.
	#
	@staticmethod
	def arpEntriesOfLocalInterfaces():
		if _has_netifaces:
			ret = []

			for i in netifaces.interfaces():
				addrs = netifaces.ifaddresses(i)
				try:
					if_mac = addrs[netifaces.AF_LINK][0]['addr']
					if_ip = addrs[netifaces.AF_INET][0]['addr']
				except IndexError as e:
					if_mac = if_ip = None
				except KeyError as e: #ignore ifaces that dont have MAC or IP
					if_mac = if_ip = None
				ret.append(ArpRecord(if_ip, -1, if_mac, None))

			return ret

		else:
			raise Exception("Python module netiface module not available!")
	#

#




















#!/usr/bin/python3




#
# This is a round-robin sequence. It internally uses a list of fixed length and insertes items in a round-robin fashion, overwriting old items as necessary.
#
class RoundRobinSequence(object):

	def __init__(self, maxLength:int):
		self.__items = [ None ] * maxLength
		self.__startI = 0
		self.__endI = 0
		self.__maxLength = maxLength
		self.__count = 0
	#

	@property
	def maxLength(self):
		return self.__maxLength
	#

	@property
	def count(self):
		return self.__count
	#

	def __len__(self):
		return self.__count
	#

	def __iter__(self):
		if self.__count > 0:
			if self.__startI < self.__endI:
				for i in range(self.__startI, self.__endI):
					yield self.__items[i]
			else:
				for i in range(self.__startI, self.__maxLength):
					yield self.__items[i]
				for i in range(0, self.__endI):
					yield self.__items[i]
	#

	def __getitem__(self, index:int):
		if self.__count:
			assert isinstance(index, int)
			n = (self.__startI + index) % self.__count
			return self.__items[n]
		else:
			raise IndexError()
	#

	def add(self, item):
		self.__items[self.__endI] = item

		self.__endI += 1
		if self.__endI == self.__maxLength:
			self.__endI = 0

		if self.__count == self.__maxLength:
			self.__startI += 1
			if self.__startI == self.__maxLength:
				self.__startI = 0
		else:
			self.__count += 1

		return self
	#

	def addRange(self, iterable):
		for item in iterable:
			self.__items[self.__endI] = item

			self.__endI += 1
			if self.__endI == self.__maxLength:
				self.__endI = 0

			if self.__count == self.__maxLength:
				self.__startI += 1
				if self.__startI == self.__maxLength:
					self.__startI = 0
			else:
				self.__count += 1

		return self
	#

	def items(self):
		if self.__count == 0:
			return []

		if self.__startI < self.__endI:
			return self.__items[self.__startI:self.__endI]
		else:
			return self.__items[self.__startI:] + self.__items[:self.__endI]
	#

	def sum(self):
		if self.__count == 0:
			return 0

		if self.__startI < self.__endI:
			return sum(self.__items[self.__startI:self.__endI])
		else:
			return sum(self.__items[self.__startI:]) + sum(self.__items[:self.__endI])
	#

	def avg(self):
		if self.__count == 0:
			return None

		if self.__startI == self.__endI:
			totalSum = sum(self.__items)
		elif self.__startI < self.__endI:
			totalSum = sum(self.__items[self.__startI:self.__endI])
		else:
			totalSum = sum(self.__items[self.__startI:]) + sum(self.__items[:self.__endI])
		return totalSum / self.__count
	#

	def min(self):
		if self.__count == 0:
			return None

		if self.__startI == self.__endI:
			return min(self.__items)
		elif self.__startI < self.__endI:
			return min(self.__items[self.__startI:self.__endI])
		else:
			a = min(self.__items[self.__startI:])
			b = min(self.__items[:self.__endI])
			return a if a < b else b
	#

	def max(self):
		if self.__count == 0:
			return None

		if self.__startI == self.__endI:
			return max(self.__items)
		elif self.__startI < self.__endI:
			return max(self.__items[self.__startI:self.__endI])
		else:
			a = max(self.__items[self.__startI:])
			if self.__endI > 0:
				b = max(self.__items[:self.__endI])
			return a if a > b else b
	#

#










from ..hex import *



class RGB(object):

	def __init__(self, r:float = 0, g:float = 0, b:float = 0, a:float = 1):
		self.r = r
		self.g = g
		self.b = b
		self.a = a
	#

	def __str__(self):
		a, r, g, b = self.toARGB256Tuple()
		return "#" + byteToHex(a) + byteToHex(r) + byteToHex(g) + byteToHex(b)
	#

	def __repr__(self):
		a, r, g, b = self.toARGB256Tuple()
		return "#" + byteToHex(a) + byteToHex(r) + byteToHex(g) + byteToHex(b)
	#

	def toHTML(self) -> str:
		r, g, b = self.toRGB256Tuple()
		return "#" + byteToHex(r) + byteToHex(g) + byteToHex(b)
	#

	def toAlphaHTML(self) -> str:
		a, r, g, b = self.toARGB256Tuple()
		return "#" + byteToHex(a) + byteToHex(r) + byteToHex(g) + byteToHex(b)
	#

	def toRGBTuple(self) -> tuple:
		return (self.r, self.g, self.b)
	#

	def toARGBTuple(self) -> tuple:
		return (self.a, self.r, self.g, self.b)
	#

	def toRGB256Tuple(self) -> tuple:
		rr = int(self.r*255)
		if rr < 0:
			rr = 0
		if rr > 255:
			rr = 255
		gg = int(self.g*255)
		if gg < 0:
			gg = 0
		if gg > 255:
			gg = 255
		bb = int(self.b*255)
		if bb < 0:
			bb = 0
		if bb > 255:
			bb = 255
		return rr, gg, bb
	#

	def toARGB256Tuple(self) -> tuple:
		aa = int(self.a*255)
		if aa < 0:
			aa = 0
		if aa > 255:
			aa = 255
		rr = int(self.r*255)
		if rr < 0:
			rr = 0
		if rr > 255:
			rr = 255
		gg = int(self.g*255)
		if gg < 0:
			gg = 0
		if gg > 255:
			gg = 255
		bb = int(self.b*255)
		if bb < 0:
			bb = 0
		if bb > 255:
			bb = 255
		return aa, rr, gg, bb
	#

	def toInt(self) -> tuple:
		a, r, g, b = self.toARGB256Tuple()
		return a * 256*256*256 + r * 256*256 + g * 256 + b
	#

	def __int__(self):
		a, r, g, b = self.toARGB256Tuple()
		return a * 256*256*256 + r * 256*256 + g * 256 + b
	#

	@staticmethod
	def parse(s:str):
		assert isinstance(s, str)
		assert s
		assert s[0] == "#"
		assert len(s) in [ 7, 9 ]
		s = s[1:]
		data = hexStrToByteArray(s)
		if len(data) == 3:
			return RGB(data[0]/255.0, data[1]/255.0, data[2]/255.0)
		else:
			return RGB(data[1]/255.0, data[2]/255.0, data[3]/255.0, data[0]/255.0)
	#

#









def create1D(n:int, value=None) -> list:
	return [ value for i in range(0, n) ]
#

def create2D(w:int, h:int, value=None) -> list:
	return [
		[ value for i in range(0, w) ]
		for j in range(0, h)
		]
#

def create3D(w:int, h:int, d:int, value=None) -> list:
	return [
			[
				[ value for i in range(0, w) ]
				for j in range(0, h)
			]
			for k in range(0, d)
		]
#









################################################################################
################################################################################
###
###  This file is automatically generated. Do not change this file! Changes
###  will get overwritten! Change the source file for "setup.py" instead.
###  This is either 'packageinfo.json' or 'packageinfo.jsonc'
###
################################################################################
################################################################################


from setuptools import setup

def readme():
	with open("README.md", "r", encoding="UTF-8-sig") as f:
		return f.read()

setup(
	author = "Jürgen Knauth",
	author_email = "pubsrc@binary-overflow.de",
	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"License :: OSI Approved :: Apache Software License",
		"Programming Language :: Python :: 3",
	],
	description = "This python module provides various utility functions and classes.",
	include_package_data = False,
	install_requires = [
		"sh",
		"sortedcontainers",
		"jk_simpleexec",
		"jk_logging",
		"jk_version",
	],
	keywords = [
		"utilities",
	],
	license = "Apache2",
	name = "jk_utils",
	packages = [
		"jk_utils",
		"jk_utils.async",
		"jk_utils.color",
		"jk_utils.datetime",
		"jk_utils.tokenizer2",
		"jk_utils.typed",
		"jk_utils.weakref",
	],
	version = "0.2022.2.6",
	zip_safe = False,
	long_description = readme(),
	long_description_content_type="text/markdown",
)






"""
from .EnumBase import EnumBase



################################################################################################################################
## class EnumNodeType
################################################################################################################################

class EnumNodeType(EnumBase):

	# The element contains a boolean value
	BOOLEAN = 10, 'boolean'
	# The element contains an integer value
	INTEGER = 20, 'integer'
	# The element contains a floatingpoint value
	FLOAT = 30, 'float'
	# The element contains a text that does not exceed 255 characters
	STRING = 40, 'string'
	# The element contains a larger amount of text (CLOB)
	TEXT = 50, 'text'
	# The element contains a timestamp (since epoch)
	TIMESTAMP = 60, 'timestamp'
	# The element contains a temperature value
	TEMPERATURE = 70, 'temperature'

	# The element contains a string list
	STRINGLIST = 140, 'stringlist'

	# The element is a new node with data.
	NODE = 200, 'node'
	# The element is a list of nodes.
	NODELIST = 210, 'nodelist'
	# The element is a map of nodes (with keys of type 'string').
	NODEMAP = 220, 'nodemap'

#




################################################################################################################################
## class DataNode
################################################################################################################################

class DataNode(dict):

	def __init__(self):
		super().__init__()

		self.description = None
	#

	def setValueIfNotNone(self, key:str, value):
		if value is None:
			return
		self[key] = value
	#

	def setIntIfNotNone(self, key:str, value):
		if value is None:
			return
		if not isinstance(value, int):
			value = int(value)
		self[key] = value
	#

	def setFloatIfNotNone(self, key:str, value):
		if value is None:
			return
		if not isinstance(value, float):
			value = float(value)
		self[key] = value
	#

	def setStrIfNotNone(self, key:str, value):
		if value is None:
			return
		if not isinstance(value, str):
			value = str(value)
		self[key] = value
	#

	def appendToListIfNotNone(self, key:str, value):
		if key in self:
			myList = self[key]
			if not isinstance(myList, list):
				raise Exception("Element at key '" + key + "' is of type '" + str(type(myList)) + "', not of type 'list'!")
			if value is None:
				return
		else:
			if value is None:
				return
			myList = []
			self[key] = myList
		myList.append(value)
	#

	def getCreateList(self, key:str):
		if key in self:
			myList = self[key]
			if not isinstance(myList, list):
				raise Exception("Element at key '" + key + "' is of type '" + str(type(myList)) + "', not of type 'list'!")
		else:
			myList = []
			self[key] = myList
		return myList
	#

#




################################################################################################################################
## class DataNodeDef
################################################################################################################################

#
# This class represents a node description: A specification of the type of data node generated.
#
class DataNodeDef(object):

	#
	# Initialization method
	#
	def __init__(self, name:str, nodeType:EnumNodeType, bMandatory:bool = False, description:str = None, children:list = None):
		assert isinstance(name, (type(None), str))
		assert isinstance(nodeType, EnumNodeType)
		assert isinstance(bMandatory, bool)
		assert isinstance(description, (type(None), str))
		assert isinstance(children, (type(None), list))

		self.__name = name
		self.__nodeType = nodeType
		self.__bMandatory = bMandatory
		self.__description = description
		self.__children = children
	#

	@property
	def children(self):
		return self.__children
	#

	@property
	def name(self):
		return self.__name
	#

	@property
	def nodeType(self):
		return self.__nodeType
	#

	@property
	def isMandatory(self):
		return self.__bMandatory
	#

	@property
	def description(self):
		return self.__description
	#

	def toJSON(self):
		ret = {
			"name": self.__name,
			"type": str(self.__nodeType),
			"mandatory": self.__bMandatory,
			"description": self.__description
		}
		if self.__children:
			jsonChildren = []
			ret["children"] = jsonChildren
			for c in self.__children:
				jsonChildren.append(c.toJSON())
		return ret
	#

	def dump(self, prefix:str=None, printFunc:callable=None):
		if prefix is None:
			prefix = ""
		if printFunc is None:
			printFunc = print

		s = "DataNodeDef( "
		if self.__name:
			s += "\"" + repr(self.__name)[1:-1] + "\""
		else:
			s += "---"
		s += ": " + str(self.__nodeType)
		if self.__bMandatory:
			s += ", mandatory"
		if self.__children:
			printFunc(prefix + s + ", children=[")
			for c in self.__children:
				c.dump(prefix + "\t", printFunc)
			printFunc(prefix + "])")
		else:
			printFunc(prefix + s + " )")
	#

	@staticmethod
	def fromJSON(jsonData:dict):
		name = jsonData.get("name", None)
		nodeType = EnumNodeType.parse(jsonData["type"])
		bMandatory = jsonData.get("mandatory", None)
		description = jsonData.get("description", None)
		children = None
		if "children" in jsonData:
			children = []
			for c in jsonData["children"]:
				children.append(DataNodeDef.fromJSON(c))
		return DataNodeDef(name, nodeType, bMandatory, description, children)
	#

#
"""




















from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# A token sequence pattern. The complete sequence must match, regardless of what kind of element it consists.
#
class TokenPatternSequence(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern[] tokenPatterns			(required) A sequence of token patterns to match against.
	#
	def __init__(self, tokenPatterns:list):
		super().__init__()

		assert isinstance(tokenPatterns, list)
		for t in tokenPatterns:
			assert isinstance(t, AbstractTokenPattern)
		self.__tokenPatterns = tokenPatterns
		self.__len = len(tokenPatterns)
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		m = stack.mark()
		orgPos = pos
		for i in range(0, self.__len):
			(bResult, n) = self.__tokenPatterns[i]._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if not bResult:
				m.resetToMark()
				return (False, 0)
			pos += n
		if pos - orgPos > 0:
			self._addTagsToStack(stack)
			return (True, pos - orgPos)
		else:
			return (False, 0)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		tokenPatterns = []
		for tokenPattern in self.__tokenPatterns:
			tokenPatterns.append(tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
		return TokenPatternSequence(tokenPatterns)
	#

#














class TabularWriterMediaWiki(object):

	def __init__(self, columns:list):
		self.__columns = columns

		print("{| class=\"wikitable\"")
		print("|-")
		print("!" + " !! ".join(columns))

		self.__proto = [ "" for x in columns ]
	#

	def print(self, column:str, text:str):
		outList = list(self.__proto)
		outList[self.__columns.index(column)] = text

		print("|-")
		print(("|" + " || ".join(outList)).replace("  ", " "))
	#

	def close(self):
		print("|}")
	#

#














class TextOutputBuffer(object):

	def __init__(self):
		self.__buf = []
	#

	@property
	def lastWasNewLine(self) -> bool:
		if self.__buf:
			return self.__buf[-1].endswith("\n")
		else:
			return True
	#

	def writeLn(self, *args):
		for a in args:
			self.__buf.append(str(a))
		self.__buf.append("\n")
	#

	def write(self, *args):
		for a in args:
			self.__buf.append(str(a))
	#

	def writeNewLines(self, n:int):
		for i in range(0, n):
			self.__buf.append("\n")
	#

	def writeNewLine(self):
		self.__buf.append("\n")
	#

	def __str__(self):
		return "".join(self.__buf)
	#

	def loadFromFile(self, filePath:str):
		with open(filePath, "r") as fin:
			self.__buf = [
				(x + "\n") for x in fin.read().split("\n")
			]
	#

	def saveToFile(self, filePath:str, encoding:str = None):
		with open(filePath, "w", encoding=encoding) as fout:
			fout.write(str(self))
	#

	@staticmethod
	def createFromFile(filePath:str):
		ret = TextOutputBuffer()
		ret.loadFromFile(filePath)
		return ret
	#

#











import sys
import warnings



__WARNING_COLOR = "\033[93m"
__RESET_COLOR = "\033[0m"
__REMEMBER_DEPRECATIONS = {}



#
# This is a decorator which can be used to mark functions as deprecated.
# It will result in a warning being emitted when the function is used.
#
def deprecated(func):

	__REMEMBER_DEPRECATIONS[func] = False

	def new_func(*args, **kwargs):
		if not __REMEMBER_DEPRECATIONS[func]:
			warnings.simplefilter("always", DeprecationWarning)  # turn off filter
			errMsg = "Call to deprecated function '{}()'.".format(func.__name__)
			if sys.stdin.isatty():
				errMsg = __WARNING_COLOR + errMsg + __RESET_COLOR
			warnings.warn(
				errMsg,
				category=DeprecationWarning,
				stacklevel=2)
			warnings.simplefilter("default", DeprecationWarning)		# reset filter
			__REMEMBER_DEPRECATIONS[func] = True

		return func(*args, **kwargs)
	#

	return new_func
#














import re
import datetime






#
# This class represents a time value with a 1 minute accuracy.
#
class T1(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self):
		self._hour = 0
		self._minute = 0
		self._absoluteTick = 0
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def hour(self) -> int:
		return self._hour
	#

	@property
	def minute(self) -> int:
		return self._minute
	#

	@property
	def hourMinute(self) -> int:
		return self._hour * 100 + self._minute
	#

	@property
	def absoluteTick(self) -> int:
		return self._absoluteTick
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def clone(self):
		ret = T1()
		ret._hour = self._hour
		ret._minute = self._minute
		ret._absoluteTick = self._absoluteTick
		return ret
	#

	################################################################################################################################
	## Special Methods
	################################################################################################################################

	def __str__(self):
		return "{:02d}:{:02d}".format(self._hour, self._minute)
	#

	def __repr__(self):
		return "{:02d}:{:02d}".format(self._hour, self._minute)
	#

	def __eq__(self, other):
		if isinstance(other, T1):
			return self._absoluteTick == other._absoluteTick
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, T1):
			return self._absoluteTick != other._absoluteTick
		else:
			return True
	#

	def __ge__(self, other):
		assert isinstance(other, T1)
		return self._absoluteTick >= other._absoluteTick
	#

	def __gt__(self, other):
		assert isinstance(other, T1)
		return self._absoluteTick > other._absoluteTick
	#

	def __le__(self, other):
		assert isinstance(other, T1)
		return self._absoluteTick <= other._absoluteTick
	#

	def __lt__(self, other):
		assert isinstance(other, T1)
		return self._absoluteTick < other._absoluteTick
	#

	def __iadd__(self, other):
		if isinstance(other, int):
			self._absoluteTick += other
		elif isinstance(other, T1):
			self._absoluteTick += other._absoluteTick
		else:
			raise Exception(str(type(other)))
		self._hour = self._absoluteTick // 60
		self._minute = self._absoluteTick % 60
		return self
	#

	def __isub__(self, other):
		if isinstance(other, int):
			self._absoluteTick -= other
		elif isinstance(other, T1):
			self._absoluteTick -= other._absoluteTick
		else:
			raise Exception(str(type(other)))
		self._hour = self._absoluteTick // 60
		self._minute = self._absoluteTick % 60
		return self
	#

	def __add__(self, other):
		if isinstance(other, int):
			return T1.createFromAbsoluteTick(self._absoluteTick + other)
		elif isinstance(other, T1):
			return T1.createFromAbsoluteTick(self._absoluteTick + other._absoluteTick)
		else:
			raise Exception(str(type(other)))
	#

	def __sub__(self, other):
		if isinstance(other, int):
			return T1.createFromAbsoluteTick(self._absoluteTick - other)
		elif isinstance(other, T1):
			return T1.createFromAbsoluteTick(self._absoluteTick - other._absoluteTick)
		else:
			raise Exception(str(type(other)))
	#

	def __int__(self):
		return self._absoluteTick
	#

	################################################################################################################################
	## Static Methods
	################################################################################################################################

	@staticmethod
	def __tryParse(s:str):
		for patternType, pattern in [
			("hms", r"(?P<hour>\d\d):(?P<minute>\d\d):(?P<second>\d\d)"),
			("hm", r"(?P<hour>\d\d):(?P<minute>\d\d)"),
		]:
			m = re.match("^" + pattern + "$", s)
			if m:
				if patternType == "hms":
					hour = int(m.group("hour"))
					minute = int(m.group("minute"))
					second = int(m.group("second"))
				elif patternType == "hm":
					hour = int(m.group("hour"))
					minute = int(m.group("minute"))
					second = 0
				else:
					raise Exception()
				if (0 <= hour <= 23) and (0 <= minute <= 59) and (0 <= second <= 59):
					yield (hour, minute, second)
	#

	@staticmethod
	def tryParseFromStr(s:str):
		if isinstance(s, str):
			s = s.strip()
			for hour, minute, second in T1.__tryParse(s):
				return T1.createFromTime(hour, minute)
		return None
	#

	@staticmethod
	def createFromTime(hour:int, minute:int):
		assert isinstance(hour, int)
		assert isinstance(minute, int)

		ret = T1()
		ret._hour = hour
		ret._minute = minute
		ret._absoluteTick = ret._hour * 60 + ret._minute
		return ret
	#

	@staticmethod
	def createFromAbsoluteTick(absoluteTick:int):
		assert isinstance(absoluteTick, int)

		ret = T1()
		ret._hour = absoluteTick // 60
		ret._minute = absoluteTick % 60
		ret._absoluteTick = absoluteTick
		return ret
	#

	@staticmethod
	def createFromHourMinutes(hourMinutes:int):
		assert isinstance(hourMinutes, int)

		ret = T1()
		ret._hour = hourMinutes // 100
		ret._minute = hourMinutes % 100
		if (ret._hour < 0) or (ret._hour > 23) or (ret._minute < 0) or (ret._minute > 59):
			raise Exception("Invalid time!")
		ret._absoluteTick = ret._hour * 60 + ret._minute
		return ret
	#

	@staticmethod
	def now():
		dt = datetime.datetime.now()

		ret = T1()
		ret._hour = dt.hour
		ret._minute = dt.minute
		ret._absoluteTick = ret._hour * 60 + ret._minute
		return ret
	#

#









#!/usr/bin/python3

import inspect
from typing import List, Union
import collections

import jk_logging





# ----------------------------------------------------------------

_ActionRecord = collections.namedtuple("__ActionRecord", [ "f", "priority", "name" ])


def _listOfActionRecordsToStr(records:list):
	ret = "[ "
	for r in records:
		if len(ret) > 2:
			ret += ", "
		ret += "(" + str(r.name if r.name else "callable") + ", " + str(r.priority) + ")"
	return ret + " ]"
#


# ----------------------------------------------------------------







class _State(object):

	def __init__(self, state):
		self.__state = state
		self.__fromStates = {}			# preconditions: int -> __ActionRecord[]
		self.__general = []				# __ActionRecord[]
	#

	def registerAction(self, fromStateOrListOfStates, action:callable, priority:int, name:str):
		assert action != None

		if fromStateOrListOfStates is None:
			self.__general.append(_ActionRecord(action, priority, name))
		else:
			if isinstance(fromStateOrListOfStates, (list, tuple)):
				for s in fromStateOrListOfStates:
					s = int(s)
					if s in self.__fromStates:
						actions = self.__fromStates[s]
					else:
						actions = []
						self.__fromStates[s] = actions
					actions.append(_ActionRecord(action, priority, name))
			else:
				s = int(fromStateOrListOfStates)
				if s in self.__fromStates:
					actions = self.__fromStates[s]
				else:
					actions = []
					self.__fromStates[s] = actions
				actions.append(_ActionRecord(action, priority, name))
	#

	def getActions(self, fromState) -> List[_ActionRecord]:
		fromState = int(fromState)
		ret = []
		ret.extend(self.__general)
		if fromState in self.__fromStates:
			ret.extend(self.__fromStates[fromState])
		return ret
	#

	def __str__(self):
		return str(self.__state)
	#

	def __repr__(self):
		return str(self.__state)
	#

	def dump(self, stateMap, prefix, outputFunction):
		outputFunction(prefix + str(self.__state) + ":")
		prefix += "\t"
		for precond in self.__fromStates:
			outputFunction(prefix + str(stateMap[precond]) + " -> " + str(self.__state) + " : " + _listOfActionRecordsToStr(self.__fromStates[precond]))
		if len(self.__general) > 0:
			outputFunction(prefix + "* -> " + str(self.__state) + " : " + _listOfActionRecordsToStr(self.__general))
	#

#

# ----------------------------------------------------------------



#
# This class manages actions on state transitions. The basic concept follows this principle:
#
# * initialization phase
#	* instantiate a state manager object; specify all possible states and the initial state;
#	* register actions with all transitions as necessary; you can choose from
#		* actions performed if a state is left
#		* actions performed if a state is reached
#		* actions performed on specific state transitions
# * runtime phase
#	* specify state transitions using <c>switchState()</c> and get actions performed automatically in sorted order: priorities of actions are considered accordingly
#
class StateManager(object):

	def __init__(self, states, startingState):
		self.__states = {}
		for state in states:
			self.__states[int(state)] = _State(state)
		self.__currentState = startingState
		self.__general = []				# __ActionRecord[]
	#

	@property
	def currentState(self) -> int:
		return self.__currentState
	#

	def registerActionFromTo(self, fromStateOrListOfStates, toStateOrListOfStates, action:callable, priority:int, name:str = None):
		assert fromStateOrListOfStates != None
		assert toStateOrListOfStates != None
		assert action != None

		if isinstance(toStateOrListOfStates, (list, tuple)):
			for s in toStateOrListOfStates:
				self.__states[int(s)].registerAction(fromStateOrListOfStates, action, priority, name)
		else:
			self.__states[int(toStateOrListOfStates)].registerAction(fromStateOrListOfStates, action, priority, name)
	#

	def registerActionTo(self, toStateOrListOfStates, action:callable, priority:int, name:str = None):
		assert toStateOrListOfStates != None
		assert action != None

		if isinstance(toStateOrListOfStates, (list, tuple)):
			for s in toStateOrListOfStates:
				self.__states[int(s)].registerAction(None, action, priority, name)
		else:
			self.__states[int(toStateOrListOfStates)].registerAction(None, action, priority, name)
	#

	def registerActionFrom(self, fromStateOrListOfStates, action:callable, priority:int, name:str = None):
		assert fromStateOrListOfStates != None
		assert action != None

		for state in self.__states.values():
			state.registerAction(fromStateOrListOfStates, action, priority, name)
	#

	def registerAction(self, action:callable, priority:int, name:str = None):
		assert action != None

		self.__general.append(_ActionRecord(action, priority, name))
	#

	#
	# Switch to a state without performaing any actions.
	#
	def switchStateWithoutActions(self, toState, logger:jk_logging.AbstractLogger = None):
		if logger != None:
			logger.debug("Setting state: " + str(toState))

		self.__currentState = toState
	#

	#
	# Switch the current state to the specified state and perform all actions defined for this transition.
	#
	def switchState(self, toState, logger:jk_logging.AbstractLogger = None):
		if self.__currentState == toState:
			return None

		if logger != None:
			logger.debug("Switching from state " + str(self.__currentState) + " to " + str(toState))

		allActions = self.__states[int(toState)].getActions(self.__currentState)
		allActions.extend(self.__general)

		# sort all actions so that actions with the highest priority come first
		allActions.sort(key=lambda a: a.priority, reverse=True)

		# remember state
		self.__currentState = toState

		# process all actions
		for a in allActions:
			if logger != None:
				logger.debug("Executing: " + str(a.f))
			a.f()
	#

	def dump(self, prefix="", outputFunction=print):
		outputFunction(prefix + "Current state: " + str(self.__currentState))
		outputFunction(prefix + "States:")
		prefix += "\t"
		for state in self.__states.values():
			state.dump(self.__states, prefix, outputFunction)
			if len(self.__general) > 0:
				outputFunction(prefix + "\t*" + " : " + _listOfActionRecordsToStr(self.__general))
	#

#


